import os
import glob


class Library:

    def __init__(self, name, paths):
        assert(isinstance(name, str)), '`name` must be a string'
        assert(isinstance(paths, list)), '`paths` must be a list'
        self.name = name
        self.paths = sum(
            [glob.glob(os.path.expanduser(p)) for p in paths],
            []
        )

    def path_format(self):
        return ":".join(self.paths)

    def __str__(self):
        return self.name


def from_paths(paths):
    name = ":".join(paths)
    return Library(name, paths)



from stevedore import extension
import logging
import re

import papis.bibtex
import papis.config
import papis.importer
import papis.plugin

logger = logging.getLogger("downloader")


downloader_mgr = None


def _create_downloader_mgr():
    global downloader_mgr

    if downloader_mgr is not None:
        return

    downloader_mgr = extension.ExtensionManager(
        namespace='papis.downloader',
        invoke_on_load=False,
        verify_requirements=True,
        propagate_map_exceptions=True,
        on_load_failure_callback=papis.plugin.stevedore_error_handler
    )


def get_available_downloaders():
    global downloader_mgr
    _create_downloader_mgr()
    return [e.plugin for e in downloader_mgr.extensions]


def get_downloader(url, downloader=''):
    """Get downloader object. If only a url is given, the url is matched
    against the match method of the downloaders.

    :param url: Url of the document
    :type  url: str
    :param downloader: Name of a downloader
    :type  downloader: str
    :returns: A Downloader if found or none
    :rtype:  papis.downloader.base.Downloader
    :raises KeyError: If no downloader is found

    """
    global downloader_mgr
    _create_downloader_mgr()
    assert(isinstance(downloader, str))
    if downloader:
        return get_downloader_by_name(downloader)(url)
    else:
        return get_matching_downloaders(url)[0]


def get_matching_downloaders(url):
    """Get matching downloaders sorted by their priorities.
    The first elements have the higher priority

    :param url: Url to be matched against
    :type  url: str
    :returns: A list of sorted downloaders
    :rtype: list
    """
    global downloader_mgr
    _create_downloader_mgr()
    return sorted(
        filter(
            lambda d: d,
            [d.match(url) for d in get_available_downloaders()]
        ),
        key=lambda k: k.priority,
        reverse=True
    )


def get_downloader_by_name(name):
    """Get a downloader by its name

    :param name: Name of the downloader
    :type  name: str
    :returns: A downloader class
    :rtype:  papis.base.Downloader
    """
    global downloader_mgr
    _create_downloader_mgr()
    return downloader_mgr[name].plugin


def get_downloaders():
    global downloader_mgr
    _create_downloader_mgr()
    return [e.plugin for e in downloader_mgr.extensions]


def get_info_from_url(url, expected_doc_format=None):
    """Get information directly from url

    :param url: Url of the resource
    :type  url: str
    :param expected_doc_format: override the doc format of the document
    :type  expected_doc_format: str
    :returns: Context object
    :rtype:  papis.importer.Context or None
    """

    downloaders = get_matching_downloaders(url)
    if not downloaders:
        logger.warning(
            "No matching downloader found for (%s)" % url
        )
        return None
    else:
        logger.debug('Found {0} matching downloaders'.format(len(downloaders)))
        downloader = downloaders[0]

    logger.info("Using downloader '%s'" % downloader)
    if downloader.expected_document_extension is None and \
            expected_doc_format is not None:
        downloader.expected_document_extension = expected_doc_format
    downloader.fetch()
    return downloader.ctx


class Importer(papis.importer.Importer):

    """Importer that tries to get data and files from implemented downloaders
    """

    def __init__(self, **kwargs):
        papis.importer.Importer.__init__(self, name='url', **kwargs)

    @classmethod
    def match(cls, uri):
        return (
            Importer(uri=uri)
            if re.match(' *http(s)?.*', uri) is not None
            else None
        )

    def fetch(self):
        self.logger.info("attempting to import from url {0}".format(self.uri))
        self.ctx = get_info_from_url(self.uri)



import re
import papis.downloaders.base
import papis.document


class Downloader(papis.downloaders.base.Downloader):

    def __init__(self, url):
        papis.downloaders.base.Downloader.__init__(self, url, name="acs")
        self.expected_document_extension = 'pdf'
        # It seems to be necessary so that acs lets us download the bibtex
        self.cookies = { 'gdpr': 'true', }
        self.priority = 10

    @classmethod
    def match(cls, url):
        return Downloader(url) if re.match(r".*acs.org.*", url) else False

    def get_data(self):
        data = dict()
        soup = self._get_soup()
        metas = soup.find_all(name="meta")
        data.setdefault('abstract', '')
        for meta in metas:
            if meta.attrs.get('name') == 'dc.Title':
                data['title'] = meta.attrs.get('content')
            elif meta.attrs.get('name') == 'keywords':
                data['keywords'] = meta.attrs.get('content')
            elif meta.attrs.get('name') == 'dc.Type':
                data['type'] = meta.attrs.get('content')
            elif meta.attrs.get('name') == 'dc.Subject':
                data['note'] = meta.attrs.get('content')
            elif (meta.attrs.get('name') == 'dc.Identifier' and
                    meta.attrs.get('scheme') == 'doi'):
                data['doi'] = meta.attrs.get('content')
            elif meta.attrs.get('name') == 'dc.Publisher':
                data['publisher'] = meta.attrs.get('content')
            elif meta.attrs.get('name') == 'dc.Description':
                data['abstract'] += meta.attrs.get('content')

        articles = soup.find_all(name='article', attrs={'class': 'article'})
        author_list = []
        if articles:
            article = articles[0]
            for author in article.find_all(name='a', attrs={'id': 'authors'}):
                author_list.append(
                    dict(
                        given=author.text.split(' ')[0],
                        family=' '.join(author.text.split(' ')[1:]),
                        affiliation=[]
                    )
                )
            year = article.find_all(
                    name='span', attrs={'class': 'citation_year'})
            if year:
                data['year'] = year[0].text
            volume = article.find_all(
                    name='span', attrs={'class': 'citation_volume'})
            if volume:
                data['volume'] = volume[0].text
            affiliations = article.find_all(
                    name='div', attrs={'class': 'affiliations'})
            if affiliations:
                # TODO: There is no guarantee that the affiliations thus
                # retrieved are ok, however is better than nothing.
                # They will most probably don't match the authors
                for aff in affiliations[0].find_all(name='div'):
                    for author in author_list:
                        author['affiliation'].append(
                            dict(name=aff.text.replace('\n', ' ')))

        data['author_list'] = author_list
        data['author'] = papis.document.author_list_to_author(data)

        return data

    def get_document_url(self):
        if 'doi' in self.ctx.data:
            return "http://pubs.acs.org/doi/pdf/" + self.ctx.data['doi']

    def get_bibtex_url(self):
        if 'doi' in self.ctx.data:
            url = ("http://pubs.acs.org/action/downloadCitation"
                  "?format=bibtex&cookieSet=1&doi=%s" % self.ctx.data['doi'])
            self.logger.debug("bibtex url = %s" % url)
            return url



import click
import papis.config
import difflib


class AliasedGroup(click.Group):
    """
    This group command is taken from
        http://click.palletsprojects.com/en/5.x/advanced/#command-aliases
    and is to be used for groups with aliases
    """

    def get_command(self, ctx, cmd_name):
        rv = click.Group.get_command(self, ctx, cmd_name)
        if rv is not None:
            return rv
        matches = difflib.get_close_matches(
            cmd_name, self.list_commands(ctx), n=2)
        if not matches:
            return None
        elif len(matches) == 1:
            return click.Group.get_command(self, ctx, matches[0])
        ctx.fail('Too many matches: %s' % ', '.join(sorted(matches)))


def query_option(**attrs):
    """Adds a ``query`` argument as a decorator"""
    def decorator(f):
        attrs.setdefault(
            'default',
            lambda: papis.config.get('default-query-string'))
        return click.decorators.argument('query', **attrs)(f)
    return decorator


def doc_folder_option(**attrs):
    """Adds a ``query`` argument as a decorator"""
    def decorator(f):
        attrs.setdefault('default', None)
        attrs.setdefault('type', click.Path(exists=True))
        attrs.setdefault('help', 'Apply action to a document path')
        return click.decorators.option('--doc-folder', **attrs)(f)
    return decorator


def git_option(help="Add git interoperability", **attrs):
    """Adds a ``git`` option as a decorator"""
    def decorator(f):
        attrs.setdefault(
            'default',
            lambda: True if papis.config.get('use-git') else False)
        attrs.setdefault('help', help)
        return click.decorators.option('--git/--no-git', **attrs)(f)
    return decorator


def bypass(group, command, command_name):
    """
    This function is specially important for people developing scripts in
    papis.

    Suppose you're writing a plugin that uses the ``add`` command as seen
    in the command line in papis. However you don't want exactly the ``add``
    command and you want to add some behavior before calling it, and you
    don't want to write your own ``add`` function from scratch.

    You can then use the following snippet

    .. code::python

        import click
        import papis.cli
        import papis.commands.add

        @click.group()
        def main():
            \"\"\"Your main app\"\"\"
            pass

        @papis.cli.bypass(main, papis.commands.add.cli, "add")
        def add(**kwargs):
            # do some logic here...
            # and call the original add command line function by
            papis.commands.add.cli.bypassed(**kwargs)
    """
    group.add_command(command, command_name)

    def decorator(new_callback):
        setattr(command, "bypassed", command.callback)
        command.callback = new_callback
    return decorator



import pickle
import logging
import os
import papis.utils
import papis.docmatcher
import papis.document
import papis.config
import papis.database.base
import re
import multiprocessing
import time
import sys


logger = logging.getLogger("cache")


def get_cache_file_name(directory):
    """Create a cache file name out of the path of a given directory.

    :param directory: Folder name to be used as a seed for the cache name.
    :type  directory: str
    :returns: Name for the cache file.
    :rtype:  str

    >>> get_cache_file_name('path/to/my/lib')
    'a8c689820a94babec20c5d6269c7d488-lib'
    >>> get_cache_file_name('papers')
    'a566b2bebc62611dff4cdaceac1a7bbd-papers'
    """
    import hashlib
    return "{}-{}".format(
        hashlib.md5(directory.encode()).hexdigest(),
        os.path.basename(directory)
    )


def get_cache_file_path(directory):
    """Get the full path to the cache file

    :param directory: Library folder
    :type  directory: str

    >>> import os; os.environ["XDG_CACHE_HOME"] = '/tmp'
    >>> get_cache_file_path('blah/papers')
    '/tmp/papis/database/c39177eca0eaea2e21134b0bd06631b6-papers'
    """
    cache_name = get_cache_file_name(directory)
    folder = os.path.expanduser(
        os.path.join(papis.utils.get_cache_home(), 'database')
    )
    if not os.path.exists(folder):
        os.makedirs(folder)
    return os.path.join(folder, cache_name)


def filter_documents(documents, search=""):
    """Filter documents. It can be done in a multi core way.

    :param documents: List of papis documents.
    :type  documents: papis.documents.Document
    :param search: Valid papis search string.
    :type  search: str
    :returns: List of filtered documents
    :rtype:  list

    >>> document = papis.document.from_data({'author': 'einstein'})
    >>> len(filter_documents([document], search="einstein")) == 1
    True
    >>> len(filter_documents([document], search="author : ein")) == 1
    True
    >>> len(filter_documents([document], search="title : ein")) == 1
    False

    """
    logger = logging.getLogger('filter')
    papis.docmatcher.DocMatcher.set_search(search)
    papis.docmatcher.DocMatcher.parse()
    papis.docmatcher.DocMatcher.set_matcher(match_document)
    begin_t = 1000 * time.time()
    # FIXME: find a better solution for this that works for both OSes
    if sys.platform == "win32":
        logger.debug(
            "Filtering {0} docs (search {1})".format(
                len(documents), search))
        filtered_docs = list(
            filter(
                lambda d: d is not None,
                map(papis.docmatcher.DocMatcher.return_if_match, documents)))
    else:
        # Doing this multiprocessing in filtering does not seem
        # to help much, I don't know if it's because I'm doing something
        # wrong or it is really like this.
        np = multiprocessing.cpu_count()
        pool = multiprocessing.Pool(np)
        logger.debug(
            "Filtering {0} docs (search {1}) using {2} cores".format(
                len(documents), search, np))
        result = pool.map(papis.docmatcher.DocMatcher.return_if_match, documents)
        pool.close()
        pool.join()
        filtered_docs = [d for d in result if d is not None]
    _delta = 1000 * time.time() - begin_t
    logger.debug("done ({0} ms) ({1} docs)".format(_delta, len(filtered_docs)))
    return filtered_docs


def match_document(document, search, match_format=None):
    """Main function to match document to a given search.

    :param document: Papis document
    :type  document: papis.document.Document
    :param search: A valid search string
    :type  search: str
    :param match_format: Python-like format string.
        (`see <
            https://docs.python.org/2/library/string.html#format-string-syntax
        >`_)
    :type  match_format: str
    :returns: Non false if matches, true-ish if it does match.

    >>> papis.config.set('match-format', '{doc[author]}')
    >>> document = papis.document.from_data({'author': 'einstein'})
    >>> match_document(document, 'e in') is None
    False
    >>> match_document(document, 'ee in') is None
    True
    >>> match_document(document, 'einstein', '{doc[title]}') is None
    True
    """
    match_format = match_format or papis.config.get("match-format")
    match_string = papis.utils.format_doc(match_format, document)
    regex = get_regex_from_search(search)
    return re.match(regex, match_string, re.IGNORECASE)


def get_regex_from_search(search):
    """Creates a default regex from a search string.

    :param search: A valid search string
    :type  search: str
    :returns: Regular expression
    :rtype: str

    >>> get_regex_from_search(' ein 192     photon')
    '.*.*ein.*192.*photon'
    """
    return r".*"+re.sub(r"\s+", ".*", search)


class Database(papis.database.base.Database):

    def __init__(self, library=None):
        papis.database.base.Database.__init__(self, library)
        self.logger = logging.getLogger('db:cache')
        self.documents = None
        self.initialize()

    def get_backend_name(self):
        return 'papis'

    def initialize(self):
        pass

    def get_documents(self):
        if self.documents is not None:
            return self.documents
        use_cache = papis.config.getboolean("use-cache")
        cache_path = self._get_cache_file_path()
        if use_cache and os.path.exists(cache_path):
            self.logger.debug(
                "Getting documents from cache in {0}".format(cache_path)
            )
            with open(cache_path, 'rb') as fd:
                self.documents = pickle.load(fd)
        else:
            self.logger.info('Indexing library, this might take a while')
            folders = sum([
                papis.utils.get_folders(d) for d in self.get_dirs()
            ], [])
            self.documents = papis.utils.folders_to_documents(folders)
            if use_cache:
                self.save()
        self.logger.debug(
            "Loaded documents ({} documents)".format(
                len(self.documents)
            )
        )
        return self.documents

    def add(self, document):
        docs = self.get_documents()
        self.logger.debug('adding ...')
        docs.append(document)
        assert(
            docs[-1].get_main_folder() == document.get_main_folder()
        )
        assert(os.path.exists(document.get_main_folder()))
        self.save()

    def update(self, document):
        if not papis.config.getboolean("use-cache"):
            return
        docs = self.get_documents()
        self.logger.debug('updating document')
        result = self._locate_document(document)
        index = result[0][0]
        docs[index] = document
        self.save()

    def delete(self, document):
        if not papis.config.getboolean("use-cache"):
            return
        docs = self.get_documents()
        self.logger.debug('deleting document')
        result = self._locate_document(document)
        index = result[0][0]
        docs.pop(index)
        self.save()

    def match(self, document, query_string):
        return match_document(document, query_string)

    def clear(self):
        cache_path = self._get_cache_file_path()
        self.logger.warning("clearing cache %s " % cache_path)
        if os.path.exists(cache_path):
            os.remove(cache_path)

    def query_dict(self, dictionary):
        query_string = " ".join(
            ["{}:\"{}\" ".format(key, val) for key, val in dictionary.items()]
        )
        return self.query(query_string)

    def query(self, query_string):
        self.logger.debug('Querying')
        docs = self.get_documents()
        # This makes it faster, if it's the all query string, return everything
        # without filtering
        if query_string == self.get_all_query_string():
            return docs
        else:
            return filter_documents(docs, query_string)

    def get_all_query_string(self):
        return '.'

    def get_all_documents(self):
        return self.get_documents()

    def save(self):
        docs = self.get_documents()
        self.logger.debug(
            'Saving ... ({} documents)'.format(len(docs))
        )
        path = self._get_cache_file_path()
        with open(path, "wb+") as fd:
            pickle.dump(docs, fd)

    def _get_cache_file_path(self):
        return get_cache_file_path(self.lib.path_format())

    def _locate_document(self, document):
        assert(isinstance(document, papis.document.Document))
        result = list(filter(
            lambda d: d[1].get_main_folder() == document.get_main_folder(),
            enumerate(self.get_documents())
        ))
        if len(result) == 0:
            raise Exception(
                'The document passed could not be found in the library'
            )
        return result



import re
import papis.downloaders.fallback


class Downloader(papis.downloaders.fallback.Downloader):

    def __init__(self, url):
        papis.downloaders.fallback.Downloader.__init__(self, url, name="hal")
        self.expected_document_extension = 'pdf'
        self.priority = 10

    @classmethod
    def match(cls, url):
        if re.match(r".*hal\.archives-ouvertes\.fr.*", url):
            return Downloader(url)
        else:
            return False

    def get_bibtex_url(self):
        if 'pdf_url' in self.ctx.data:
            url = re.sub(r'document', 'bibtex', self.uri)
            self.logger.debug('bibtex url = {url}'.format(url=url))
            return url



import papis
import logging
import os.path
from stevedore import extension
import papis.plugin

logger = logging.getLogger('importer')


class Context:
    def __init__(self):
        self.data = dict()
        self.files = []

    def __bool__(self):
        return bool(self.files) or bool(self.data)


class Importer:

    """This is the base class for every importer"""

    def __init__(self, uri="", name="", ctx=None):
        """
        :param uri: uri
        :type  uri: str
        :param name: Name of the importer
        :type  name: str
        """
        self.ctx = ctx or Context()
        assert(isinstance(uri, str))
        assert(isinstance(name, str))
        assert(isinstance(self.ctx, Context))
        self.uri = uri
        self.name = name or os.path.basename(__file__)
        self.logger = logging.getLogger("importer:{0}".format(self.name))

    @classmethod
    def match(uri):
        """This method should be called to know if a given uri matches
        the importer or not.

        For example, a valid match for archive would be:
        .. code:: python

            return re.match(r".*arxiv.org.*", uri)

        it will return something that is true if it matches and something
        falsely otherwise.

        :param uri: uri where the document should be retrieved from.
        :type  uri: str
        """
        raise NotImplementedError(
            "Matching uri not implemented for this importer"
        )

    @classmethod
    def match_data(cls, data):
        """Get a dictionary of data and try to decide if there is
        a valid uri in it.

        :param data: Data to look into
        :type  data: dict
        """
        raise NotImplementedError(
            "Matching data not implemented for this importer"
        )

    def fetch(self):
        """
        can return a dict to update the document with
        """
        raise NotImplementedError()

    def __str__(self):
        return 'Importer({0}, uri={1})'.format(self.name, self.uri)


import_mgr = None


def _create_import_mgr():
    global import_mgr
    if import_mgr:
        return
    import_mgr = extension.ExtensionManager(
        namespace='papis.importer',
        invoke_on_load=True,
        verify_requirements=True,
        invoke_args=(),
        # invoke_kwds
        propagate_map_exceptions=True,
        on_load_failure_callback=papis.plugin.stevedore_error_handler
    )


def get_import_mgr():
    """Get the import manager
    :returns: Import manager
    :rtype:  stevedore.extension.ExtensionManager
    """
    global import_mgr
    _create_import_mgr()
    return import_mgr


def available_importers():
    """Get the available importers defined.
    :returns: List of importer names
    :rtype:  list(str)
    """
    return get_import_mgr().entry_points_names()


def get_importers():
    return [e.plugin for e in get_import_mgr().extensions]


def get_importer_by_name(name):
    """Get importer by name
    :param name: Name of the importer
    :type  name: str
    :returns: The importer
    :rtype:  Importer
    """
    assert(isinstance(name, str))
    return get_import_mgr()[name].plugin


def cache(f):
    """
    This is a decorator to be used if a method of an Importer
    is to be cached, i.e., if the context of the importer is already
    set, then one does not need to run the function anymore, even if
    it is explicitly run.

    :param self: Method of an Importer
    """
    def wrapper(self):
        if not self.ctx:
            f(self)
    return wrapper



import re
import papis.downloaders.base


class Downloader(papis.downloaders.base.Downloader):

    def __init__(self, url):
        papis.downloaders.base.Downloader.__init__(
            self, url, name="annualreviews"
        )
        self.expected_document_extension = 'pdf'
        self.priority = 10

    @classmethod
    def match(cls, url):
        if re.match(r".*annualreviews.org.*", url):
            return Downloader(url)
        else:
            return False

    def get_document_url(self):
        if 'doi' in self.ctx.data:
            doi = self.ctx.data['doi']
            url = "http://annualreviews.org/doi/pdf/%s" % doi
            self.logger.debug("doc url = %s" % url)
            return url

    def get_bibtex_url(self):
        if 'doi' in self.ctx.data:
            url = ("http://annualreviews.org/action/downloadCitation"
                   "?format=bibtex&cookieSet=1&doi=%s" % self.ctx.data['doi'])
            self.logger.debug("bibtex url = %s" % url)
            return url

    def get_data(self):
        data = dict()
        soup = self._get_soup()
        data.update(papis.downloaders.base.parse_meta_headers(soup))

        if 'author_list' in data:
            return data

        # Read brute force the authors from the source
        author_list = []
        authors = soup.find_all(name='span', attrs={'class': 'contribDegrees'})
        cleanregex = re.compile(r'(^\s*|\s*$|&)')
        editorregex = re.compile(r'([\n|]|\(Reviewing\s*Editor\))')
        morespace = re.compile(r'\s+')
        for author in authors:
            affspan = author.find_all('span', attrs={'class': 'overlay'})
            afftext = affspan[0].text if affspan else ''
            fullname = re.sub(',', '',
                          cleanregex.sub('',
                              author.text.replace(afftext, '')))
            splitted = re.split(r'\s+', fullname)
            cafftext = re.sub(' ,', ',',
                              morespace.sub(' ', cleanregex.sub('', afftext)))
            if 'Reviewing Editor' in fullname:
                data['editor'] = cleanregex.sub(
                    ' ', editorregex.sub('', fullname))
                continue
            given = splitted[0]
            family = ' '.join(splitted[1:])
            author_list.append(
                dict(
                    given=given,
                    family=family,
                    affiliation=[dict(name=cafftext)] if cafftext else []
                )
            )

        data['author_list'] = author_list
        data['author'] = papis.document.author_list_to_author(data)

        return data



import os
import re
import shutil
import logging

import papis.utils
import papis.config
import papis.bibtex


def keyconversion_to_data(key_conversion, data, keep_unknown_keys=False):
    new_data = dict()
    for orig_key in key_conversion:
        if orig_key not in data:
            continue

        conv_data_list = key_conversion[orig_key]
        if isinstance(conv_data_list, dict):
            conv_data_list = [conv_data_list]

        for conv_data in conv_data_list:
            papis_key = conv_data.get('key', orig_key)
            papis_value = data[orig_key]

            try:
                action = conv_data.get('action', lambda x: x)
                new_data[papis_key] = action(papis_value)
            except Exception as e:
                logger.debug(
                    "Error while trying to parse {0} ({1})".format(
                        papis_key, e))

    if keep_unknown_keys:
        for key, value in data.items():
            if key in key_conversion:
                continue
            new_data[key] = value

    if 'author_list' in new_data:
        new_data['author'] = author_list_to_author(new_data)

    return new_data


def author_list_to_author(data):
    author = ''
    if 'author_list' in data:
        author = (
            papis.config.get('multiple-authors-separator')
            .join([
                papis.config.get("multiple-authors-format").format(au=author)
                for author in data['author_list']
            ])
        )
    return author


class Author(dict):
    """Base class for authors, if you're parsing an author,
    use this class.
    """
    def __init__(self, given, family, affiliations=[]):
        dict.__init__(
            self, given=given, family=family, affiliations=affiliations
        )


class Affiliation(dict):
    """Base class for affiliation, if you're parsing an affiliations,
    use this class."""
    def __init__(self, name):
        dict.__init__(self, name=name)


logger = logging.getLogger("document")


def new(folder_path, data, files=[]):
    """
    Creates a document at a given folder with data and
    some existing files.

    :param folder_path: A folder path, if non existing it will be created
    :type  folder_path: str
    :param data: Dictionary with key and values to be updated
    :type  data: dict
    :param files: Existing paths for files
    :type  files: list(str)
    :raises FileExistsError: If folder_path exists
    """
    assert(isinstance(folder_path, str))
    assert(isinstance(data, dict))
    assert(isinstance(files, list))
    os.makedirs(folder_path)
    doc = Document(folder=folder_path, data=data)
    doc['files'] = []
    for f in files:
        shutil.copy(f, os.path.join(folder_path))
        doc['files'].append(os.path.basename(f))
    doc.save()
    return doc


def from_folder(folder_path):
    """Construct a document object from a folder

    :param folder_path: Full path to a valid papis folder
    :type  folder_path: str
    :returns: A papis document
    :rtype:  papis.document.Document
    """
    return papis.document.Document(folder=folder_path)


def from_data(data):
    """Construct a document object from a data dictionary.

    :param data: Data to be copied to a new document
    :type  data: dict
    :returns: A papis document
    :rtype:  papis.document.Document
    """
    return papis.document.Document(data=data)


def to_bibtex(document):
    """Create a bibtex string from document's information

    :param document: Papis document
    :type  document: Document
    :returns: String containing bibtex formating
    :rtype:  str

    """
    logger = logging.getLogger("document:bibtex")
    bibtexString = ""
    bibtexType = ""

    # First the type, article ....
    if "type" in document.keys():
        if document["type"] in papis.bibtex.bibtex_types:
            bibtexType = document["type"]
        elif document["type"] in papis.bibtex.bibtex_type_converter.keys():
            bibtexType = papis.bibtex.bibtex_type_converter[document["type"]]
    if not bibtexType:
        bibtexType = "article"

    # REFERENCE BUILDING
    if document.has("ref"):
        ref = document["ref"]
    elif papis.config.get('ref-format'):
        try:
            ref = papis.utils.format_doc(
                papis.config.get("ref-format"),
                document
            ).replace(" ", "")
        except Exception as e:
            logger.error(e)
            ref = None

    logger.debug("generated ref=%s" % ref)
    if not ref:
        if document.has('doi'):
            ref = document['doi']
        else:
            try:
                ref = os.path.basename(document.get_main_folder())
            except:
                ref = 'noreference'

    ref = re.sub(r'[;,()\/{}\[\]]', '', ref)
    logger.debug("Used ref=%s" % ref)

    bibtexString += "@{type}{{{ref},\n".format(type=bibtexType, ref=ref)
    for bibKey in sorted(document.keys()):
        logger.debug('%s : %s' % (bibKey, document[bibKey]))
        if bibKey in papis.bibtex.bibtex_key_converter:
            newBibKey = papis.bibtex.bibtex_key_converter[bibKey]
            document[newBibKey] = document[bibKey]
            continue
        if bibKey in papis.bibtex.bibtex_keys:
            value = str(document[bibKey])
            if not papis.config.get('bibtex-unicode'):
                value = papis.bibtex.unicode_to_latex(value)
            if bibKey == 'journal':
                bibtex_journal_key = papis.config.get('bibtex-journal-key')
                if bibtex_journal_key in document.keys():
                    bibtexString += "  %s = {%s},\n" % (
                        'journal',
                        papis.bibtex.unicode_to_latex(
                            str(
                              document[papis.config.get('bibtex-journal-key')]
                            )
                        )
                    )
                elif bibtex_journal_key not in document.keys():
                    logger.warning(
                        "Key '%s' is not present for ref=%s" % (
                            papis.config.get('bibtex-journal-key'),
                            document["ref"]
                        )
                    )
                    bibtexString += "  %s = {%s},\n" % (
                        'journal',
                        value
                    )
            else:
                bibtexString += "  %s = {%s},\n" % (
                    bibKey,
                    value
                )
    bibtexString += "}\n"
    return bibtexString


def to_json(document):
    """Export information into a json string
    :param document: Papis document
    :type  document: Document
    :returns: Json formatted info file
    :rtype:  str

    """
    import json
    return json.dumps(to_dict(document))


def to_dict(document):
    """Gets a python dictionary with the information of the document
    :param document: Papis document
    :type  document: Document
    :returns: Python dictionary
    :rtype:  dict
    """
    result = dict()
    for key in document.keys():
        result[key] = document[key]
    return result


def dump(document):
    """Return information string without any obvious format
    :param document: Papis document
    :type  document: Document
    :returns: String with document's information
    :rtype:  str

    >>> doc = from_data({'title': 'Hello World'})
    >>> dump(doc)
    'title:   Hello World\\n'
    """
    string = ""
    for i in document.keys():
        string += str(i)+":   "+str(document[i])+"\n"
    return string


def delete(document):
    """This function deletes a document from disk.
    :param document: Papis document
    :type  document: papis.document.Document
    """
    folder = document.get_main_folder()
    shutil.rmtree(folder)


def describe(document):
    """Return a string description of the current document
    using the document-description-format
    """
    return papis.utils.format_doc(
        papis.config.get('document-description-format'),
        document
    )


def move(document, path):
    """This function moves a document to path, it supposes that
    the document exists in the location ``document.get_main_folder()``.
    Warning: This method will change the folder in the document object too.
    :param document: Papis document
    :type  document: papis.document.Document
    :param path: Full path where the document will be moved to
    :type  path: str

    >>> doc = from_data({'title': 'Hello World'})
    >>> doc.set_folder('path/to/folder')
    >>> import tempfile; newfolder = tempfile.mkdtemp()
    >>> move(doc, newfolder)
    Traceback (most recent call last):
    ...
    Exception: There is already...
    """
    path = os.path.expanduser(path)
    if os.path.exists(path):
        raise Exception(
            "There is already a document in {0}, please check it,\n"
            "a temporary papis document has been stored in {1}".format(
                path, document.get_main_folder()
            )
        )
    shutil.move(document.get_main_folder(), path)
    # Let us chmod it because it might come from a temp folder
    # and temp folders are per default 0o600
    os.chmod(path, papis.config.getint('dir-umask'))
    document.set_folder(path)


class DocHtmlEscaped(dict):
    """
    Small helper class to escape html elements.

    >>> DocHtmlEscaped(from_data(dict(title='> >< int & "" "')))['title']
    '&gt; &gt;&lt; int &amp; &quot;&quot; &quot;'
    """

    def __init__(self, doc):
        self.__doc = doc

    def __getitem__(self, key):
        return (
            str(self.__doc[key])
            .replace('&', '&amp;')
            .replace('<', '&lt;')
            .replace('>', '&gt;')
            .replace('"', '&quot;')
        )


class Document(dict):

    """Class implementing the entry abstraction of a document in a library.
    It is basically a python dictionary with more methods.
    """

    subfolder = ""
    _info_file_path = ""

    def __init__(self, folder=None, data=None):
        self._folder = None

        if folder is not None:
            self.set_folder(folder)
            self.load()

        if data is not None:
            self.update(data)

    def __missing__(self, key):
        """
        If key is not defined, return empty string
        """
        return ""

    @property
    def html_escape(self):
        return DocHtmlEscaped(self)

    def get_main_folder(self):
        """Get full path for the folder where the document and the information
        is stored.
        :returns: Folder path
        """
        return self._folder

    def set_folder(self, folder):
        """Set document's folder. The info_file path will be accordingly set.

        :param folder: Folder where the document will be stored, full path.
        :type  folder: str
        """
        self._folder = folder
        self._info_file_path = os.path.join(
            folder,
            papis.config.get('info-name')
        )
        # TODO: check if this makes sense at all
        self.subfolder = self.get_main_folder().replace(
            os.path.expanduser("~"), ""
        ).replace(
            "/", " "
        )

    def get_main_folder_name(self):
        """Get main folder name where the document and the information is
        stored.
        :returns: Folder name
        """
        return os.path.basename(self._folder)

    def has(self, key):
        """Check if the information file has some key defined.

        :param key: Key name to be checked
        :returns: True/False

        """
        return key in self

    def save(self):
        """Saves the current document's information into the info file.
        """
        import papis.yaml
        papis.yaml.data_to_yaml(
            self.get_info_file(),
            {key: self[key] for key in self.keys() if self[key]}
        )

    def get_info_file(self):
        """Get full path for the info file
        :returns: Full path for the info file
        :rtype: str
        """
        return self._info_file_path

    def get_files(self):
        """Get the files linked to the document, if any.

        :returns: List of full file paths
        :rtype:  list
        """
        result = []
        if not self.has('files'):
            return result
        files = (
            self["files"] if isinstance(self["files"], list)
            else [self["files"]]
        )
        for f in files:
            result.append(os.path.join(self.get_main_folder(), f))
        return result

    def load(self):
        """Load information from info file
        """
        import papis.yaml
        if not os.path.exists(self.get_info_file()):
            return
        try:
            data = papis.yaml.yaml_to_data(
                self.get_info_file(), raise_exception=True)
        except Exception as e:
            logger.error(
                'Error reading yaml file in {0}'.format(yaml_path) +
                '\nPlease check it!\n\n{0}'.format(str(e))
            )
        else:
            for key in data:
                self[key] = data[key]



import os
import logging
import requests
import papis.config
import papis.document
import papis.utils
import filetype
import papis.importer
import papis.bibtex
import tempfile
import copy
import re
import bs4


meta_equivalences = [
# google
{"tag": "meta", "key": "abstract", "attrs": {"name": "description"}},
{"tag": "meta", "key": "doi", "attrs": {"name": "doi"}},
{"tag": "meta", "key": "keywords", "attrs": {"name": "keywords"}},
{"tag": "title", "key": "title", "attrs": {}, "action": lambda e: e.text},
# facebook
{"tag": "meta", "key": "type", "attrs": {"property": "og:type"}},
{"tag": "meta", "key": "abstract", "attrs": {"property": "og:description"}},
{"tag": "meta", "key": "title", "attrs": {"property": "og:title"}},
{"tag": "meta", "key": "url", "attrs": {"property": "og:url"}},
# citation style
# https://scholar.google.com/intl/en/scholar/inclusion.html#indexing
{"tag": "meta", "key": "doi", "attrs": {"name": "citation_doi"}},
{"tag": "meta", "key": "firstpage", "attrs": {"name": "citation_firstpage"}},
{"tag": "meta", "key": "lastpage", "attrs": {"name": "citation_lastpage"}},
{"tag": "meta", "key": "url", "attrs": {"name": "citation_fulltext_html_url"}},
{"tag": "meta", "key": "pdf_url", "attrs": {"name": "citation_pdf_url"}},
{"tag": "meta", "key": "issn", "attrs": {"name": "citation_issn"}},
{"tag": "meta", "key": "issue", "attrs": {"name": "citation_issue"}},
{"tag": "meta", "key": "abstract", "attrs": {"name": "citation_abstract"}},
{"tag": "meta", "key": "journal_abbrev", "attrs": {"name": "citation_journal_abbrev"}},
{"tag": "meta", "key": "journal", "attrs": {"name": "citation_journal_title"}},
{"tag": "meta", "key": "language", "attrs": {"name": "citation_language"}},
{"tag": "meta", "key": "online_date", "attrs": {"name": "citation_online_date"}},
{"tag": "meta", "key": "publication_date", "attrs": {"name": "citation_publication_date"}},
{"tag": "meta", "key": "publisher", "attrs": {"name": "citation_publisher"}},
{"tag": "meta", "key": "title", "attrs": {"name": "citation_title"}},
{"tag": "meta", "key": "volume", "attrs": {"name": "citation_volume"}},
# dc.{id} style
{"tag": "meta", "key": "publisher", "attrs": {"name": re.compile("dc.publisher", re.I)}},
{"tag": "meta", "key": "publisher", "attrs": {"name": re.compile(".*st.publisher.*", re.I)}},
{"tag": "meta", "key": "date", "attrs": {"name": re.compile("dc.date", re.I)}},
{"tag": "meta", "key": "language", "attrs": {"name": re.compile("dc.language", re.I)}},
{"tag": "meta", "key": "issue", "attrs": {"name": re.compile("dc.citation.issue", re.I)}},
{"tag": "meta", "key": "volume", "attrs": {"name": re.compile("dc.citation.volume", re.I)}},
{"tag": "meta", "key": "keywords", "attrs": {"name": re.compile("dc.subject", re.I)}},
{"tag": "meta", "key": "title", "attrs": {"name": re.compile("dc.title", re.I)}},
{"tag": "meta", "key": "type", "attrs": {"name": re.compile("dc.type", re.I)}},
{"tag": "meta", "key": "abstract", "attrs": {"name": re.compile("dc.description", re.I)}},
{"tag": "meta", "key": "abstract", "attrs": {"name": re.compile("dc.description.abstract", re.I)}},
{"tag": "meta", "key": "journal_abbrev", "attrs": {"name": re.compile("dc.relation.ispartof", re.I)}},
{"tag": "meta", "key": "year", "attrs": {"name": re.compile("dc.issued", re.I)}},
{"tag": "meta", "key": "doi", "attrs": {"name": re.compile("dc.identifier", re.I), "scheme": "doi"}},
]


def parse_meta_headers(soup, extra_equivalences=[]):
    equivalences = copy.copy(meta_equivalences)
    equivalences.extend(extra_equivalences)
    #metas = soup.find_all(name="meta")
    data = dict()
    for equiv in equivalences:
        elements = soup.find_all(equiv['tag'], attrs=equiv["attrs"])
        if elements:
            if "action" in equiv:
                value = equiv["action"](elements[0])
            else:
                value = elements[0].attrs.get("content")
            if isinstance(value, str):
                data[equiv["key"]] = value.replace('\r', '')
            else:
                data[equiv["key"]] = value

    author_list = parse_meta_authors(soup)
    if author_list:
        data['author_list'] = author_list
        data['author'] = papis.document.author_list_to_author(data)

    return data


def parse_meta_authors(soup):
    author_list = []
    authors = soup.find_all(name='meta', attrs={'name': 'citation_author'})
    if not authors:
        authors = soup.find_all(
            name='meta', attrs={'name': re.compile('dc.creator', re.I)})
    affs = soup.find_all(name='meta',
            attrs={'name': 'citation_author_institution'})
    if affs and authors:
        tuples = zip(authors, affs)
    elif authors:
        tuples = [(a, None) for a in authors]
    else:
        return []

    for t in tuples:
        fullname = t[0].get('content')
        affiliation = [dict(name=t[1].get('content'))] if t[1] else []
        fullnames = re.split('\s+', fullname)
        author_list.append(dict(
            given=fullnames[0],
            family=' '.join(fullnames[1:]),
            affiliation=affiliation,
        ))
    return author_list


class Downloader(papis.importer.Importer):

    """This is the base class for every downloader.
    """

    def __init__(self, uri="", name="", ctx=None):
        self.ctx = ctx or papis.importer.Context()
        assert(isinstance(uri, str))
        assert(isinstance(name, str))
        assert(isinstance(self.ctx, papis.importer.Context))
        self.uri = uri
        self.name = name or os.path.basename(__file__)
        self.logger = logging.getLogger("downloader:"+self.name)
        self.logger.debug("uri {0}".format(uri))
        self.expected_document_extension = None
        self.priority = 1
        self._soup = None

        self.bibtex_data = None
        self.document_data = None

        self.session = requests.Session()
        self.session.headers = {
            'User-Agent': papis.config.get('user-agent')
        }
        proxy = papis.config.get('downloader-proxy')
        if proxy is not None:
            self.session.proxies = {
                'http': proxy,
                'https': proxy,
            }
        self.cookies = {}

    def fetch_data(self):
        """
        Try first to get data by hand with the get_data command.
        Then commplement with bibtex data.
        """
        # Try with get_data
        try:
            data = self.get_data()
            assert(isinstance(data, dict))
        except NotImplementedError:
            pass
        else:
            self.ctx.data.update(data)

        # try with bibtex
        try:
            self.download_bibtex()
        except NotImplementedError:
            pass
        else:
            bib_rawdata = self.get_bibtex_data()
            if bib_rawdata:
                datalist = papis.bibtex.bibtex_to_dict(bib_rawdata)
                if datalist:
                    self.logger.info("Merging data from bibtex")
                    self.ctx.data.update(datalist[0])
        # try getting doi
        try:
            doi = self.get_doi()
        except NotImplementedError:
            pass
        else:
            self.ctx.data['doi'] = doi


    def fetch_files(self):
        # get documents
        try:
            self.download_document()
        except NotImplementedError:
            pass
        else:
            doc_rawdata = self.get_document_data()
            if doc_rawdata and self.check_document_format():
                tmp = tempfile.mktemp()
                self.logger.info("Saving downloaded file in {0}".format(tmp))
                with open(tmp, 'wb+') as fd:
                    fd.write(doc_rawdata)
                self.ctx.files.append(tmp)

    def fetch(self):
        self.fetch_data()
        self.fetch_files()

    def _get_body(self):
        """Get body of the uri, this is also important for unittesting"""
        return self.session.get(self.uri).content

    def _get_soup(self):
        """Get body of the uri, this is also important for unittesting"""
        if self._soup:
            return self._soup
        self._soup = bs4.BeautifulSoup(self._get_body(), features='lxml')
        return self._soup

    def __str__(self):
        return 'Downloader({0}, uri={1})'.format(self.name, self.uri)

    def get_bibtex_url(self):
        """It returns the urls that is to be access to download
        the bibtex information. It has to be implemented for every
        downloader, or otherwise it will raise an exception.

        :returns: Bibtex url
        :rtype:  str
        """
        raise NotImplementedError(
            "Getting bibtex url not implemented for this downloader"
        )

    def get_bibtex_data(self):
        """Get the bibtex_data data if it has been downloaded already
        and if not download it and return the data in utf-8 format.

        :returns: Bibtex data in utf-8 format
        :rtype:  str
        """
        if not self.bibtex_data:
            self.download_bibtex()
        return self.bibtex_data

    def download_bibtex(self):
        """Bibtex downloader, it should try to download bibtex information from
        the url provided by ``get_bibtex_url``.

        It sets the ``bibtex_data`` attribute if it succeeds.

        :returns: Nothing
        :rtype:  None
        """
        url = self.get_bibtex_url()
        if not url:
            return False
        res = self.session.get(url, cookies=self.cookies)
        self.logger.info("downloading bibtex from {0}".format(url))
        self.bibtex_data = res.content.decode()

    def get_document_url(self):
        """It returns the urls that is to be access to download
        the document information. It has to be implemented for every
        downloader, or otherwise it will raise an exception.

        :returns: Document url
        :rtype:  str
        """
        raise NotImplementedError(
            "Getting bibtex url not implemented for this downloader"
        )

    def get_data(self):
        """A general data retriever, for instance when data needn't need
        to come from a bibtex
        """
        raise NotImplementedError(
            "Getting general data is not implemented for this downloader"
        )

    def get_doi(self):
        """It returns the doi of the document, if it is retrievable.
        It has to be implemented for every downloader, or otherwise it will
        raise an exception.

        :returns: Document doi
        :rtype:  str
        """
        raise NotImplementedError(
            "Getting document url not implemented for this downloader"
        )

    def get_document_data(self):
        """Get the document_data data if it has been downloaded already
        and if not download it and return the data in binary format.

        :returns: Document data in binary format
        :rtype:  str
        """
        if not self.document_data:
            self.download_document()
        return self.document_data

    def download_document(self):
        """Document downloader, it should try to download document information from
        the url provided by ``get_document_url``.

        It sets the ``document_data`` attribute if it succeeds.

        :returns: Nothing
        :rtype:  None
        """
        url = self.get_document_url()
        if not url:
            return False
        self.logger.info("downloading file from {0}".format(url))
        res = self.session.get(url, cookies=self.cookies)
        self.document_data = res.content

    def check_document_format(self):
        """Check if the downloaded document has the filetype that the
        downloader expects. If the downloader does not expect any special
        filetype, accept anything because there is no way to know if it is
        correct.

        :returns: True if it is of the right type, else otherwise
        :rtype:  bool
        """
        def print_warning():
            self.logger.error(
                "The downloaded data does not seem to be of"
                "the correct type (%s)" % self.expected_document_extension
            )

        if self.expected_document_extension is None:
            return True

        retrieved_kind = filetype.guess(self.get_document_data())

        if retrieved_kind is None:
            print_warning()
            return False

        self.logger.debug(
            'retrieved kind of document seems to be {0}'.format(
                retrieved_kind.mime)
        )

        if not isinstance(self.expected_document_extension, list):
            expected_document_extensions = [
                self.expected_document_extension
            ]

        if retrieved_kind.extension in expected_document_extensions:
            return True
        else:
            print_warning()
            return False



"""
This command is to list contents of a library.

CLI Examples
^^^^^^^^^^^^

- List all document files associated will all entries:

    .. code:: bash

        papis list --all --file

    .. raw:: HTML

        <script type="text/javascript"
        src="https://asciinema.org/a/XwD0ZaUORoOonwDw4rXoQDkjZ.js"
        id="asciicast-XwD0ZaUORoOonwDw4rXoQDkjZ" async></script>

- List all document year and title with custom formatting:

    .. code:: bash

        papis list --all --format '{doc[year]} {doc[title]}'

    .. raw:: HTML

        <script type="text/javascript"
        src="https://asciinema.org/a/NZ8Ii1wWYPo477CIL4vZhUqOy.js"
        id="asciicast-NZ8Ii1wWYPo477CIL4vZhUqOy" async></script>

- List all documents according to the bibitem formatting (stored in a template
  file ``bibitem.template``):

    .. code:: bash

        papis list --all --template bibitem.template

    .. raw:: HTML

        <script type="text/javascript"
        src="https://asciinema.org/a/QZTBZ3tFfyk9WQuJ9WWB2UpSw.js"
        id="asciicast-QZTBZ3tFfyk9WQuJ9WWB2UpSw" async></script>

Cli
^^^
.. click:: papis.commands.list:cli
    :prog: papis list
"""

import logging
import papis
import os
import papis.utils
import papis.strings
import papis.config
import papis.database
import papis.downloaders
import papis.cli
import papis.pick
import click

logger = logging.getLogger('list')


def run(
        documents,
        libraries=False,
        downloaders=False,
        pick=False,
        files=False,
        folders=False,
        info_files=False,
        notes=False,
        fmt="",
        template=None
        ):
    """Main method to the list command

    :returns: List different objects
    :rtype:  list
    """
    if downloaders:
        return papis.downloaders.get_available_downloaders()

    if template is not None:
        if not os.path.exists(template):
            logger.error("Template file %s not found".format(template))
            return
        with open(template) as fd:
            fmt = fd.read()

    if libraries:
        config = papis.config.get_configuration()
        return [
            section + ' ' + config[section]['dir']
            for section in config
            if 'dir' in config[section] ]

    if files:
        return [
            doc_file for files in [
                document.get_files() for document in documents
            ] for doc_file in files
        ]
    elif notes:
        return [
            os.path.join(d.get_main_folder(), d["notes"])
            for d in documents
            if d.has("notes") and isinstance(d["notes"], str)
                and os.path.exists(
                        os.path.join(d.get_main_folder(), d["notes"]))]
    elif info_files:
        return [
            os.path.join(
                document.get_main_folder(),
                document.get_info_file()
            ) for document in documents
        ]
    elif fmt:
        return [
            papis.utils.format_doc(fmt, document)
            for document in documents
        ]
    elif folders:
        return [
            document.get_main_folder() for document in documents
        ]
    else:
        return documents


@click.command("list")
@click.help_option('--help', '-h')
@papis.cli.query_option()
@click.option(
    "-i",
    "--info",
    help="Show the info file name associated with the document",
    default=False,
    is_flag=True)
@click.option(
    "-f", "--file", "_file",
    help="Show the file name associated with the document",
    default=False,
    is_flag=True)
@click.option(
    "-d", "--dir", "_dir",
    help="Show the folder name associated with the document",
    default=False,
    is_flag=True)
@click.option(
    "-n", "--notes",
    help="List notes files, if any",
    default=False,
    is_flag=True)
@click.option(
    "--format", "_format",
    help="List entries using a custom papis format, e.g."
    " '{doc[year] {doc[title]}",
    default=None)
@click.option(
    "--template",
    help="Template file containing a papis format to list entries",
    default=None)
@click.option(
    "--downloaders",
    help="List available downloaders",
    default=False,
    is_flag=True)
@click.option(
    "--libraries",
    help="List defined libraries",
    default=False,
    is_flag=True)
@click.option(
    "-a", "--all", "_all",
    help="Process all documents matching a query, if a query is given",
    default=False,
    is_flag=True)
def cli(query, info, _file, notes, _dir, _format,
        template, _all, downloaders, libraries,):
    """List documents' properties"""

    logger = logging.getLogger('cli:list')
    documents = []

    if not libraries and not downloaders and not _file and not info and not _dir:
        _dir = True

    if not libraries and not downloaders:
        db = papis.database.get()
        documents = db.query(query)
        if not documents:
            logger.warning(papis.strings.no_documents_retrieved_message)
        if not _all:
            documents = filter(lambda x: x, [papis.pick.pick_doc(documents)])

    objects = run(
        documents,
        libraries=libraries,
        downloaders=downloaders,
        notes=notes,
        files=_file,
        folders=_dir,
        info_files=info,
        fmt=_format,
        template=template)

    for o in objects:
        click.echo(o)



import papis.importer
import papis.downloaders.base


class Downloader(papis.downloaders.base.Downloader):

    def __init__(self, url):
        papis.downloaders.base.Downloader.__init__(
            self, uri=url, name="pubmed")

    def get_bibtex_url(self):
        return (
            "http://pubmed.macropus.org/articles/"
            "?format=text%2Fbibtex&id={pmid}"
            .format(pmid=self.uri)
        )


class Importer(papis.importer.Importer):

    """Importer downloading data from a pubmed id"""

    def __init__(self, uri='', **kwargs):
        papis.importer.Importer.__init__(
            self, name='pubmed', uri=uri, **kwargs)
        self.downloader = Downloader(uri)

    @classmethod
    def match(cls, uri):
        # TODO:
        pass

    def fetch(self):
        self.downloader.fetch()
        self.ctx = self.downloader.ctx



"""
The add command is one of the central commands of the papis command line
interface. It is a very versatile command with a fair amount of options.

There are also customization settings availabe for this command, check out
the :ref:`configuration page <add-command-options>` for add.

Examples
^^^^^^^^

- Add a document located in ``~/Documents/interesting.pdf`` and name the
  folder where it will be stored in the database ``interesting-paper-2021``

    .. code::

        papis add ~/Documents/interesting.pdf \\
            --folder-name interesting-paper-2021

  if you want to add directly some key values, like ``author``, ``title``
  and ``tags``, you can also run the following:

    .. code::

        papis add ~/Documents/interesting.pdf \\
            --folder-name interesting-paper-2021 \\
            --set author 'John Smith' \\
            --set title 'The interesting life of bees' \\
            --set year 1985 \\
            --set tags 'biology interesting bees'

- Add a paper that you have locally in a file and get the paper information
  through its ``doi`` identifier (in this case ``10.10763/1.3237134`` as an
  example):

    .. code::

        papis add ~/Documents/interesting.pdf --from doi 10.10763/1.3237134

- Add paper to a library named ``machine-learning`` from ``arxiv.org``

    .. code::

        papis -l machine-learning add \\
            --from url https://arxiv.org/abs/1712.03134

- If you do not want copy the original pdfs into the library, you can
  also tell papis to just create a link to them, for example

    .. code::

        papis add --link ~/Documents/interesting.pdf \\
            --from doi 10.10763/1.3237134

  will add an entry into the papis library, but the pdf document will remain
  at ``~/Documents/interesting.pdf``, and in the document's folder
  there will be a link to ``~/Documents/interesting.pdf`` instead of the
  file itself. Of course you always have to be sure that the
  document at ``~/Documents/interesting.pdf`` does not disappear, otherwise
  you will end up without a document to open.

- Papis also tries to make sense of the inputs that you have passed
  to the command, for instance you could provide only a ``doi`` and
  papis will try to know if this is indeed a ``doi``

    .. code::

        papis add 10.1103/PhysRevLett.123.156401

  or from a ``url``

    .. code::

        papis add https://journals.aps.org/prl/abstract/10.1103/PhysRevLett.123.156401


Examples in python
^^^^^^^^^^^^^^^^^^

There is a python function in the add module that can be used to interact
in a more effective way in python scripts,

.. autofunction:: papis.commands.add.run

Cli
^^^
.. click:: papis.commands.add:cli
    :prog: papis add

"""
import logging
from string import ascii_lowercase
import os
import re
import tempfile
import hashlib
import shutil
import papis.api
import papis.pick
import papis.utils
import papis.config
import papis.document
import papis.importer
import papis.cli
import click
import colorama
import papis.downloaders
import papis.git

logger = logging.getLogger('add')


class FromFolderImporter(papis.importer.Importer):

    """Importer that gets files and data from a valid papis folder"""

    def __init__(self, **kwargs):
        papis.importer.Importer.__init__(self, name='folder', **kwargs)

    @classmethod
    def match(cls, uri):
        return FromFolderImporter(uri=uri) if os.path.isdir(uri) else None

    def fetch(self):
        doc = papis.document.from_folder(self.uri)
        self.logger.info('importing from folder {0}'.format(self.uri))
        self.ctx.data = papis.document.to_dict(doc)
        self.ctx.files = doc.get_files()


class FromLibImporter(papis.importer.Importer):

    """Importer that queries a valid papis library (also paths) and adds files
    and data
    """

    def __init__(self, **kwargs):
        papis.importer.Importer.__init__(self, name='lib', **kwargs)

    @classmethod
    def match(cls, uri):
        try:
            papis.config.get_lib_from_name(uri)
        except Exception:
            return None
        else:
            return FromLibImporter(uri=uri)

    def fetch(self):
        doc = papis.pick.pick_doc(papis.api.get_all_documents_in_lib(self.uri))
        importer = FromFolderImporter(uri=doc.get_main_folder())
        importer.fetch()
        self.ctx = importer.ctx


def get_file_name(data, original_filepath, suffix=""):
    """Generates file name for the document

    :param data: Data parsed for the actual document
    :type  data: dict
    :param original_filepath: The full path to the original file
    :type  original_filepath: str
    :param suffix: Possible suffix to be appended to the file without
        its extension.
    :type  suffix: str
    :returns: New file name
    :rtype:  str

    """

    basename_limit = 150
    file_name_opt = papis.config.get('add-file-name')
    ext = papis.utils.get_document_extension(original_filepath)

    if file_name_opt is None:
        file_name_opt = os.path.basename(original_filepath)

    # Get a file name from the format `add-file-name`
    file_name_base = papis.utils.format_doc(
        file_name_opt,
        papis.document.from_data(data)
    )

    if len(file_name_base) > basename_limit:
        logger.warning(
            "Shortening the name {0} for portability".format(file_name_base)
        )
        file_name_base = file_name_base[0:basename_limit]

    # Remove extension from file_name_base, if any
    file_name_base = re.sub(
        r"([.]{0})?$".format(ext),
        '',
        file_name_base
    )

    # Adding some extra suffixes, if any, and cleaning up document name
    filename_basename = papis.utils.clean_document_name(
        "{}{}".format(
            file_name_base,
            "-" + suffix if len(suffix) > 0 else ""
        )
    )

    # Adding the recognised extension
    filename = filename_basename + '.' + ext

    return filename


def get_hash_folder(data, document_paths):
    """Folder name where the document will be stored.

    :data: Data parsed for the actual document
    :document_paths: Path of the document

    """
    import random
    author = "-{:.20}".format(data["author"])\
             if "author" in data.keys() else ""

    document_strings = b''
    for docpath in document_paths:
        with open(docpath, 'rb') as fd:
            document_strings += fd.read(2000)

    md5 = hashlib.md5(
        ''.join(document_paths).encode() +
        str(data).encode() +
        str(random.random()).encode() +
        document_strings
    ).hexdigest()

    result = md5 + author
    result = papis.utils.clean_document_name(result)

    return result


def run(
        paths,
        data=dict(),
        folder_name=None,
        file_name=None,
        subfolder=None,
        confirm=False,
        open_file=False,
        edit=False,
        git=False,
        link=False
        ):
    """
    :param paths: Paths to the documents to be added
    :type  paths: []
    :param data: Data for the document to be added.
        If more data is to be retrieved from other sources, the data dictionary
        will be updated from these sources.
    :type  data: dict
    :param folder_name: Name of the folder where the document will be stored
    :type  folder_name: str
    :param file_name: File name of the document's files to be stored.
    :type  file_name: str
    :param subfolder: Folder within the library where the document's folder
        should be stored.
    :type  subfolder: str
    :param confirm: Wether or not to ask user for confirmation before adding.
    :type  confirm: bool
    :param open_file: Wether or not to ask user for opening file before adding.
    :type  open_file: bool
    :param edit: Wether or not to ask user for editing the infor file
        before adding.
    :type  edit: bool
    :param git: Wether or not to ask user for committing before adding,
        in the case of course that the library is a git repository.
    :type  git: bool
    """

    logger = logging.getLogger('add:run')
    # The folder name of the new document that will be created
    out_folder_name = None
    # The real paths of the documents to be added
    in_documents_paths = paths
    # The basenames of the documents to be added
    in_documents_names = []
    # The folder name of the temporary document to be created
    temp_dir = tempfile.mkdtemp()

    for p in in_documents_paths:
        if not os.path.exists(p):
            raise IOError('Document %s not found' % p)

    in_documents_names = [
        papis.utils.clean_document_name(doc_path)
        for doc_path in in_documents_paths
    ]

    tmp_document = papis.document.Document(temp_dir)

    if not folder_name:
        out_folder_name = get_hash_folder(data, in_documents_paths)
        logger.info("Got an automatic folder name")
    else:
        temp_doc = papis.document.Document(data=data)
        out_folder_name = papis.utils.format_doc(folder_name, temp_doc)
        out_folder_name = papis.utils.clean_document_name(out_folder_name)
        del temp_doc

    data["files"] = in_documents_names
    out_folder_path = os.path.expanduser(
        os.path.join(
            papis.config.get_lib_dirs()[0],
            subfolder or '',
            out_folder_name))

    logger.info("The folder name is {0}".format(out_folder_name))
    logger.debug("Folder path: {0}".format(out_folder_path))
    logger.debug("File(s): {0}".format(in_documents_paths))

    # First prepare everything in the temporary directory
    g = papis.utils.create_identifier(ascii_lowercase)
    string_append = ''
    if file_name is not None:  # Use args if set
        papis.config.set("add-file-name", file_name)
    new_file_list = []

    for in_file_path in in_documents_paths:

        # Rename the file in the staging area
        new_filename = papis.utils.clean_document_name(
            get_file_name(
                data,
                in_file_path,
                suffix=string_append))
        new_file_list.append(new_filename)

        tmp_end_filepath = os.path.join(
            tmp_document.get_main_folder(),
            new_filename)
        string_append = next(g)

        if link:
            in_file_abspath = os.path.abspath(in_file_path)
            logger.debug(
                "[SYMLINK] '%s' to '%s'" %
                (in_file_abspath, tmp_end_filepath))
            os.symlink(in_file_abspath, tmp_end_filepath)
        else:
            logger.debug(
                "[CP] '%s' to '%s'" %
                (in_file_path, tmp_end_filepath))
            shutil.copy(in_file_path, tmp_end_filepath)

    data['files'] = new_file_list
    tmp_document.update(data)
    tmp_document.save()

    # Check if the user wants to edit before submitting the doc
    # to the library
    if edit:
        logger.info("Editing file before adding it")
        papis.api.edit_file(tmp_document.get_info_file(), wait=True)
        logger.info("Loading the changes made by editing")
        tmp_document.load()

    # Duplication checking
    logger.info("Checking if this document is already in the library")
    try:
        found_document = papis.utils.locate_document_in_lib(tmp_document)
    except IndexError:
        logger.info("No document matching found already in the library")
    else:
        logger.warning(
            colorama.Fore.RED +
            "DUPLICATION WARNING" +
            colorama.Style.RESET_ALL)
        logger.warning(
            "The document beneath is in your library and it seems to match")
        logger.warning(
            "the one that you're trying to add, "
            "I will prompt you for confirmation")
        logger.warning(
            "(Hint) Use the update command if you just want to update"
            " the info.")
        papis.utils.text_area(
            'The following document is already in your library',
            papis.document.dump(found_document),
            lexer_name='yaml',
            height=20)
        confirm = True

    if open_file:
        for d_path in tmp_document.get_files():
            papis.utils.open_file(d_path)
    if confirm:
        if not papis.utils.confirm('Really add?'):
            return

    logger.info(
        "[MV] '%s' to '%s'" %
        (tmp_document.get_main_folder(), out_folder_path)
    )
    # This also sets the folder of tmp_document
    papis.document.move(tmp_document, out_folder_path)
    papis.database.get().add(tmp_document)
    if git:
        papis.git.add_and_commit_resource(
            tmp_document.get_main_folder(), '.',
            "Add document '{0}'".format(papis.document.describe(tmp_document)))


@click.command(
    "add",
    help="Add a document into a given library"
)
@click.help_option('--help', '-h')
@click.argument("files", type=str, nargs=-1)
@click.option(
    "-s", "--set", "set_list",
    help="Set some information before",
    multiple=True,
    type=(str, str))
@click.option(
    "-d", "--subfolder",
    help="Subfolder in the library",
    default="")
@click.option(
    "--folder-name",
    help="Name for the document's folder (papis format)",
    default=lambda: papis.config.get('add-folder-name'))
@click.option(
    "--file-name",
    help="File name for the document (papis format)",
    default=None)
@click.option(
    "--from", "from_importer",
    help="Add document from a specific importer ({0})".format(
        ", ".join(papis.importer.available_importers())
    ),
    type=(click.Choice(papis.importer.available_importers()), str),
    nargs=2,
    multiple=True,
    default=(),)
@click.option(
    "-b", "--batch",
    help="Batch mode, do not prompt or otherwise",
    default=False, is_flag=True)
@click.option(
    "--confirm/--no-confirm",
    help="Ask to confirm before adding to the collection",
    default=lambda: True if papis.config.get('add-confirm') else False)
@click.option(
    "--open/--no-open", "open_file",
    help="Open file before adding document",
    default=lambda: True if papis.config.get('add-open') else False)
@click.option(
    "--edit/--no-edit",
    help="Edit info file before adding document",
    default=lambda: True if papis.config.get('add-edit') else False)
@click.option(
    "--link/--no-link",
    help="Instead of copying the file to the library, create a link to"
         "its original location",
    default=False)
@papis.cli.git_option(help="Git add and commit the new document")
@click.option(
    "--list-importers", "--li", "list_importers",
    help="List all available papis importers",
    default=False,
    is_flag=True)
def cli(files, set_list, subfolder, folder_name, file_name, from_importer,
        batch, confirm, open_file, edit, git, link, list_importers):

    if list_importers:
        import_mgr = papis.importer.get_import_mgr()
        for n in import_mgr.names():
            print("{name}\n\t{text}".format(
                name=n,
                text=re.sub(r"[ \n]+", " ", import_mgr[n].plugin.__doc__)))
        return

    from_importer = list(from_importer)
    logger = logging.getLogger('cli:add')

    data = dict()
    for data_set in set_list:
        data[data_set[0]] = data_set[1]

    files = list(files)
    ctx = papis.importer.Context()
    ctx.files = [f for f in files if os.path.exists(f)]
    ctx.data.update(data)

    if batch:
        edit = False
        confirm = False
        open_file = False

    import_mgr = papis.importer.get_import_mgr()
    matching_importers = []

    if not from_importer and not batch and files:
        matching_importers = sum((
            papis.utils.get_matching_importer_or_downloader(f)
            for f in files), [])

    for importer_tuple in from_importer:
        try:
            importer_name = importer_tuple[0]
            resource = importer_tuple[1]
            importer = import_mgr[importer_name].plugin(uri=resource)
            importer.fetch()
            if importer.ctx:
                matching_importers.append(importer)
        except Exception as e:
            logger.exception(e)

    if matching_importers:
        logger.info(
            'There are {0} possible matchings'.format(len(matching_importers)))

        for importer in matching_importers:
            if importer.ctx.data:
                logger.info(
                    'Merging data from importer {0}'.format(importer.name))
                if batch:
                    ctx.data.update(importer.ctx.data)
                else:
                    papis.utils.update_doc_from_data_interactively(
                        ctx.data,
                        importer.ctx.data,
                        str(importer))
            if importer.ctx.files:
                logger.info(
                    'Got files {0} from importer {1}'
                    .format(importer.ctx.files, importer.name))
                for f in importer.ctx.files:
                    papis.utils.open_file(f)
                    if batch or papis.utils.confirm("Use this file?"):
                        ctx.files.append(f)

    if not ctx:
        logger.error('there is nothing to be added')
        return

    return run(
        ctx.files,
        data=ctx.data,
        folder_name=folder_name,
        file_name=file_name,
        subfolder=subfolder,
        confirm=confirm,
        open_file=open_file,
        edit=edit,
        git=git,
        link=link)



import os
import glob
from stevedore import extension
import logging
import papis.config
import papis.plugin
import re


commands_mgr = None


def _create_commands_mgr():
    global commands_mgr

    if commands_mgr is not None:
        return

    commands_mgr = extension.ExtensionManager(
        namespace='papis.command',
        invoke_on_load=False,
        verify_requirements=True,
        propagate_map_exceptions=True,
        on_load_failure_callback=papis.plugin.stevedore_error_handler
    )


def get_external_scripts():
    regex = re.compile('.*papis-([^ .]+)$')
    paths = []
    scripts = {}
    paths.append(papis.config.get_scripts_folder())
    paths += os.environ["PATH"].split(":")
    for path in paths:
        for script in glob.glob(os.path.join(path, "papis-*")):
            m = regex.match(script)
            if m is not None:
                name = m.group(1)
                scripts[name] = dict(
                    command_name=name,
                    path=script,
                    plugin=None
                )
    return scripts


def get_scripts():
    global commands_mgr
    _create_commands_mgr()
    scripts_dict = dict()
    for command_name in commands_mgr.names():
        scripts_dict[command_name] = dict(
            command_name=command_name,
            path=None,
            plugin=commands_mgr[command_name].plugin
        )
    return scripts_dict



import logging

logger = logging.getLogger("plugin")


def stevedore_error_handler(manager, entrypoint, exception):
    logger.error("Error while loading entrypoint [%s]" % entrypoint)
    logger.error(exception)



import re
import papis.downloaders.base


class Downloader(papis.downloaders.base.Downloader):

    def __init__(self, url):
        papis.downloaders.base.Downloader.__init__(
            self, url, name="iopscience"
        )
        self.expected_document_extension = 'pdf'
        self.priority = 10

    @classmethod
    def match(cls, url):
        url = re.sub(r'/pdf', '', url)
        if re.match(r".*iopscience\.iop\.org.*", url):
            return Downloader(url)
        else:
            return False

    def get_doi(self):
        return self.ctx.data.get('doi')

    def get_document_url(self):
        if 'pdf_url' in self.ctx.data:
            return self.ctx.data.get('pdf_url')
        doi = self.get_doi()
        if doi:
            durl = 'https://iopscience.iop.org/article/{0}/pdf'.format(doi)
            self.logger.debug("doc url = %s" % durl)
            return durl

    def _get_article_id(self):
        """Get article's id for IOP
        :returns: Article id
        """
        doi = self.get_doi()
        if doi:
            articleId = doi.replace('10.1088/', '')
            self.logger.debug("articleId = %s" % articleId)
            return articleId

    def get_bibtex_url(self):
        aid = self._get_article_id()
        if aid:
            url = "{0}{1}{2}".format(
                "http://iopscience.iop.org/export?aid=",
                aid,
                "&exportFormat=iopexport_bib&exportType=abs"
                "&navsubmit=Export%2Babstract"
            )
            self.logger.debug("bibtex url = %s" % url)
            return url

    def get_data(self):
        data = dict()
        soup = self._get_soup()
        data.update(papis.downloaders.base.parse_meta_headers(soup))
        abstract_nodes = soup.find_all(
                'div', attrs={'class': 'wd-jnl-art-abstract'})
        if abstract_nodes:
            data['abstract'] = ' '.join(a.text for a in abstract_nodes)
        return data



import subprocess
import os
import shlex
import logging

logger = logging.getLogger("papis.git")


def _issue_git_command(path, cmd):
    """Issues a general git command ``cmd`` at ``path``.

    :param path: Folder where a git repo exists.
    :type  path: str
    :param message: Git command
    :type  message: str
    :returns: None

    """
    global logger
    assert(type(cmd) == str)
    path = os.path.expanduser(path)
    assert(os.path.exists(path))
    cmd = shlex.split(cmd)
    os.chdir(path)
    logger.debug(cmd)
    subprocess.call(cmd)


def commit(path, message):
    """Commits changes in the path with a message.

    :param path: Folder where a git repo exists.
    :type  path: str
    :param message: Commit message
    :type  message: str
    :returns: None

    """
    global logger
    logger.info('Commiting {path} with message {message}'.format(**locals()))
    cmd = 'git commit -m "{0}"'.format(message)
    _issue_git_command(path, cmd)


def add(path, resource):
    """Adds changes in the path with a message.

    :param path: Folder where a git repo exists.
    :type  path: str
    :param resource: Commit resource
    :type  resource: str
    :returns: None

    """
    global logger
    logger.info('Adding {path}'.format(**locals()))
    cmd = 'git add "{0}"'.format(resource)
    _issue_git_command(path, cmd)


def rm(path, resource, recursive=False):
    """Adds changes in the path with a message.

    :param path: Folder where a git repo exists.
    :type  path: str
    :param resource: Commit resource
    :type  resource: str
    :returns: None

    """
    global logger
    logger.info('Removing {path}'.format(**locals()))
    # force removal always
    cmd = 'git rm -f {r} "{0}"'.format(resource, r="-r" if recursive else "")
    _issue_git_command(path, cmd)



def add_and_commit_resource(path, resource, message):
    add(path, resource)
    commit(path, message)



import doi
import papis.downloaders.base


class Downloader(papis.downloaders.base.Downloader):

    def __init__(self, url, name="fallback"):
        papis.downloaders.base.Downloader.__init__(self, url, name=name)
        self.priority = -1

    @classmethod
    def match(cls, url):
        return Downloader(url)

    def get_data(self):
        data = dict()
        soup = self._get_soup()
        data.update(papis.downloaders.base.parse_meta_headers(soup))
        return data

    def get_doi(self):
        if 'doi' in self.ctx.data:
            return self.ctx.data['doi']
        soup = self._get_soup()
        self.logger.info('trying to parse doi...')
        return doi.find_doi_in_text(str(soup))

    def get_document_url(self):
        if 'pdf_url' in self.ctx.data:
            url = self.ctx.data.get('pdf_url')
            self.logger.debug("got a pdf url = %s" % url)
            return url



"""This command will try its best to find a source in the internet for the
document at hand.

Of course if the document has an url key in its info file, it will use this url
to open it in a browser.  Also if it has a doc_url key, or a doi, it will try
to compose urls out of these to open it.

If none of the above work, then it will try to use a search engine with the
document's information (using the ``browse-query-format``).  You can select
wich search engine you want to use using the ``search-engine`` setting.

Cli
^^^
.. click:: papis.commands.browse:cli
    :prog: papis browse
"""
import papis
import papis.utils
import papis.config
import papis.cli
import papis.pick
import click
import papis.database
import papis.strings
import papis.document
from urllib.parse import urlencode
import logging


logger = logging.getLogger('browse')


def run(document):
    """Browse document's url whenever possible.

    :document: Document object
    :returns: Returns the url that is composed from the document
    :rtype:  str

    """
    global logger
    url = None
    key = papis.config.get("browse-key")

    if document.has(key):
        if "doi" == key:
            url = 'https://doi.org/{}'.format(document['doi'])
        elif "isbn" == key:
            url = 'https://isbnsearch.org/isbn/{}'.format(document['isbn'])
        else:
            url = document[key]

    if url is None or key == 'search-engine':
        params = {
            'q': papis.utils.format_doc(
                papis.config.get('browse-query-format'),
                document
            )
        }
        url = papis.config.get('search-engine') + '/?' + urlencode(params)

    logger.info("Opening url %s:" % url)
    papis.utils.general_open(url, "browser", wait=False)
    return url


@click.command("browse")
@click.help_option('--help', '-h')
@papis.cli.query_option()
@click.option(
    '-k', '--key', default='',
    help='Use the value of the document\'s key to open in the browser, e.g.'
         'doi, url, doc_url ...'
)
@click.option(
    '--all', default=False, is_flag=True,
    help='Browse all selected documents'
)
def cli(query, key, all):
    """Open document's url in a browser"""
    documents = papis.database.get().query(query)
    logger = logging.getLogger('cli:browse')

    if len(documents) == 0:
        logger.warning(papis.strings.no_documents_retrieved_message)
        return 0

    if not all:
        document = papis.pick.pick_doc(documents)
        if not document:
            return
        documents = [document]

    if len(key):
        papis.config.set('browse-key', key)

    for document in documents:
        run(document)



import re
import papis.downloaders.base
import papis.document
import json
import collections
import functools


def get_author_list(data):
    rdata = []
    assert(isinstance(data, list))
    for d in data:
        if d["#name"] == "author":
            author_data = dict()
            for prop in d["$$"]:
                if prop["#name"] == "given-name":
                    author_data["given"] = prop["_"]
                elif prop["#name"] == "surname":
                    author_data["family"] = prop["_"]
                elif prop["#name"] == "cross-ref":
                    author_data["refid"] = prop["$"]["refid"]
            rdata.append(author_data)
        if d["#name"] == "affiliation":
            affid = d["$"]["id"]
            text = functools.reduce(lambda x, y: x + y,
                map(lambda x: x["_"],
                    filter(lambda x: x["#name"] == "textfn",
                        d["$$"])))
            authors = list(filter(lambda a: a.get("refid") == affid, rdata))
            if authors:
                authors[0]["affiliation"] = [dict(name=text)]
                del authors[0]["refid"]
            else:
                if len(rdata) == 1:
                    rdata[0]["affiliation"] = [dict(name=text)]
    return rdata


script_keyconv = collections.OrderedDict({
    "article": {
        "key": "_article_data",
        "action": lambda x:
            papis.document.keyconversion_to_data(article_keyconv, x)
    },
    "abstracts": {
        "key": "_abstract_data",
        "action": lambda x:
            papis.document.keyconversion_to_data(abstracts_keyconv, x)
    },
    "authors": {
        "key": "_author_data",
        "action": lambda x:
            papis.document.keyconversion_to_data(authors_keyconv, x)
    },
})
authors_keyconv = collections.OrderedDict({
    "content": {
        "key": "author_list",
        "action": lambda x:
        get_author_list(
            list(
                functools.reduce(lambda s, t: s+t,
                    map(lambda s: s["$$"],
                    filter(lambda s: s["#name"] == "author-group",
                        x)))))
    }
})
abstracts_keyconv = collections.OrderedDict({
    "content": [
        { # Single author format
            "key": "abstract",
            "action": lambda x: " ".join(
                map(lambda s: s["_"],
                    functools.reduce(lambda s, t: s + t,
                        map(lambda s: s["$$"],
                            filter(lambda s: s["#name"] == 'abstract-sec',
                                functools.reduce(lambda s, t: s+t,
                                    map(lambda s: s["$$"],
                                        filter(lambda s:
                                            s['$']['class'] == 'author',
                                            x)))))))
            )
        }, { # multiple author format (apparently)
            "key": "abstract",
            "action": lambda x: " ".join(
                map(lambda s: s['_'],
                    functools.reduce(lambda x, y: x + y,
                        map(lambda s: s["$$"],
                            filter(lambda s: s['#name'] == 'simple-para',
                                functools.reduce(lambda x, y: x + y,
                                    map(lambda s: s["$$"],
                                        filter(lambda s: s["#name"] == 'abstract-sec',
                                            functools.reduce(lambda x, y: x + y,
                                                map(lambda s: s["$$"],
                                                    filter(lambda s: s['$']['class'] == 'author',
                                                        x)))))))))))
        }
    ]
})
article_keyconv = collections.OrderedDict({
    "doi": {},
    "pii": {},
    "language": {},
    "subtitle": {},
    "issn": {},
    "srctitle": {"key": "journal"},
    "vol-first": {"key": "volume"},
    "cover-date-years": {"key": "year", "action": lambda x: x[0]},
    "cover-date-start": {"key": "date"},
    "document-type": {"key": "type"},
    "titleString": {"key": "title"},
    "dates": [
        {"key": "accepted-date", "action": lambda x: x["Accepted"]},
        {"key": "publication-date", "action": lambda x: x["Publication date"]}
    ],
    "pages": {
        "action": lambda x: "{x[first-page]}--{x[last-page]}".format(x=x[0])
    },
})


class Downloader(papis.downloaders.base.Downloader):

    def __init__(self, url):
        papis.downloaders.base.Downloader.__init__(
            self, url, name="sciencedirect"
        )
        self.expected_document_extension = 'pdf'

    @classmethod
    def match(cls, url):
        if re.match(r".*\.sciencedirect\.com.*", url):
            return Downloader(url)
        else:
            return False

    def get_data(self):
        data = dict()
        soup = self._get_soup()
        scripts = soup.find_all(name="script", attrs={'data-iso-key': '_0'})
        if scripts:
            rawdata = json.loads(scripts[0].text)
            self.logger.debug(
                "found {0} scripts data-iso-key".format(len(scripts)))
            converted_data = papis.document.keyconversion_to_data(
                script_keyconv, rawdata)
            data.update(converted_data["_article_data"])
            data.update(converted_data["_abstract_data"])
            data.update(converted_data["_author_data"])
            data['author'] = papis.document.author_list_to_author(data)
        return data



import re
import urllib.parse

import papis.downloaders.base
import papis.document


class Downloader(papis.downloaders.base.Downloader):
    re_clean = re.compile(r'(^\s*|\s*$|\s{2,}?|&)')
    re_comma = re.compile(r'(\s*,\s*)')
    re_add_dot = re.compile(r'(\b\w\b)')

    def __init__(self, url):
        papis.downloaders.base.Downloader.__init__(
            self, url, name="tandfonline")
        self.expected_document_extension = 'pdf'
        self.priority = 10

    @classmethod
    def match(cls, url):
        return (Downloader(url)
                if re.match(r".*tandfonline.com.*", url) else False)

    def get_data(self):
        data = dict()
        soup = self._get_soup()
        data.update(papis.downloaders.base.parse_meta_headers(soup))

        # `author` and `author_list` are already in meta_headers, but we
        # brute-force them here again to get exact affiliation information
        author_list = []
        authors = soup.find_all(name='span', attrs={'class': 'contribDegrees'})
        for author in authors:
            affiliation = author.find_all('span', attrs={'class': 'overlay'})
            if affiliation:
                # the span contains other things like the email, but we only
                # want the starting text with the affiliation address
                affiliation = next(affiliation[0].children)

                affiliation = self.re_comma.sub(', ',
                        self.re_clean.sub('', affiliation))
                affiliation = [dict(name=affiliation)]

            # find href="/author/escaped_fullname"
            fullname = author.find_all('a', attrs={'class': 'entryAuthor'})
            fullname = fullname[0].get('href').split('/')[-1]

            fullname = urllib.parse.unquote_plus(fullname)
            family, given = re.split(r',\s+', fullname)
            given = self.re_add_dot.sub(r'\1.', given)

            if 'Reviewing Editor' in author.text:
                data['editor'] = \
                    papis.config.get('multiple-authors-format').format(
                            au=dict(family=family, given=given))
                continue

            new_author = dict(given=given, family=family)
            if affiliation:
                new_author['affiliation'] = affiliation
            author_list.append(new_author)

        data['author_list'] = author_list
        data['author'] = papis.document.author_list_to_author(data)

        return data

    def get_bibtex_url(self):
        if 'doi' in self.ctx.data:
            url = (
                "http://www.tandfonline.com/action/downloadCitation"
                "?format=bibtex&cookieSet=1&doi=%s" % self.ctx.data['doi'])
            self.logger.debug("bibtex url = %s" % url)
            return url

    def get_document_url(self):
        if 'doi' in self.ctx.data:
            durl = (
                "http://www.tandfonline.com/doi/pdf/{doi}"
                .format(doi=self.ctx.data['doi']))
            self.logger.debug("doc url = %s" % durl)
            return durl



"""
The export command is useful to work with other programs such as bibtex.

Some examples of its usage are:

- Export one of the documents matching the author with einstein to bibtex:

.. code::

    papis export --format bibtex 'author : einstein'

or export all of them

.. code::

    papis export --format bibtex --all 'author : einstein'

- Export all documents to bibtex and save them into a ``lib.bib`` file

.. code::

    papis export --all --format bibtex --out lib.bib

- Export a folder of one of the documents matching the word ``krebs``
  into a folder named, ``interesting-document``

.. code::

    papis export --folder --out interesting-document krebs

  this will create the folder ``interesting-document`` containing the
  ``info.yaml`` file, the linked documents and a ``bibtex`` file for
  sharing with other people.


Cli
^^^
.. click:: papis.commands.export:cli
    :prog: papis export
"""
import papis
import os
import shutil
import papis.utils
import papis.document
import click
import papis.cli
import papis.api
import papis.database
import papis.strings
import logging
from stevedore import extension
import papis.plugin

logger = logging.getLogger('cli:export')
exporters_mgr = None


def export_to_yaml(documents):
    import yaml
    return yaml.dump_all(
        [
            papis.document.to_dict(document) for document in documents
        ],
        allow_unicode=True
    )


def export_to_json(documents):
    import json
    return json.dumps(
        [
            papis.document.to_dict(document) for document in documents
        ]
    )


def export_to_bibtex(documents):
    return '\n'.join([
        papis.document.to_bibtex(document) for document in documents
    ])


def available_formats():
    global exporters_mgr
    _create_mgr()
    return exporters_mgr.entry_points_names()



def _create_mgr():
    global exporters_mgr
    if exporters_mgr:
        return
    exporters_mgr = extension.ExtensionManager(
        namespace='papis.exporter',
        invoke_on_load=False,
        verify_requirements=True,
        propagate_map_exceptions=True,
        on_load_failure_callback=papis.plugin.stevedore_error_handler
    )


def run(
    documents,
    to_format,
):
    """
    Exports several documents into something else.

    :param document: A ist of papis document
    :type  document: [papis.document.Document]
    :param to_format: what format to use
    :type  to_format: str
    """
    global exporters_mgr
    _create_mgr()
    try:
        ret_string = exporters_mgr[to_format].plugin(
            document for document in documents
        )
        return ret_string
    except KeyError as e:
        logger.error("Format %s not supported." % to_format)

    return None


@click.command("export")
@click.help_option('--help', '-h')
@papis.cli.query_option()
@click.option(
    "--folder",
    help="Export document folder to share",
    default=False,
    is_flag=True
)
@click.option(
    "-o",
    "--out",
    help="Outfile or outdir",
    default=None
)
@click.option(
    "-f",
    "--format",
    help="Format for the document",
    type=click.Choice(available_formats()),
    default="bibtex",
)
@click.option(
    "-a", "--all",
    help="Export all without picking",
    default=False,
    is_flag=True
)
def cli(
        query,
        folder,
        out,
        format,
        all,
        **kwargs
        ):
    """Export a document from a given library"""

    documents = papis.database.get().query(query)

    if format and folder:
        logger.warning("Only --folder flag will be considered")

    if not documents:
        logger.warning(papis.strings.no_documents_retrieved_message)
        return

    if not all:
        document = papis.pick.pick_doc(documents)
        if not document:
            return 0
        documents = [document]

    ret_string = run(documents, to_format=format)

    if ret_string is not None and not folder:
        if out is not None:
            logger.info("Dumping to {0}".format(out))
            with open(out, 'a+') as fd:
                fd.write(ret_string)
        else:
            logger.info("Dumping to stdout")
            print(ret_string)
        return 0

    for document in documents:
        if folder:
            folder = document.get_main_folder()
            outdir = out or document.get_main_folder_name()
            if not len(documents) == 1:
                outdir = os.path.join(
                    (out or ''), document.get_main_folder_name()
                )
            logger.info("Exporting doc {0} to {1}".format(
                papis.document.describe(document), outdir
            ))
            shutil.copytree(folder, outdir)


@click.command('export')
@click.pass_context
@click.help_option('--help', '-h')
@click.option(
    "-f",
    "--format",
    help="Format for the document",
    type=click.Choice(available_formats()),
    default="bibtex",
)
@click.option(
    "-o",
    "--out",
    help="Outfile to write information to",
    type=click.Path(),
    default=None,
)
def explorer(ctx, format, out):
    """
    Export retrieved documents into various formats for later use

    Examples of its usage are

    papis explore crossref -m 200 -a 'Schrodinger' export --yaml lib.yaml

    """
    logger = logging.getLogger('explore:yaml')
    docs = ctx.obj['documents']

    outstring = run(docs, to_format=format)
    if out is not None:
        with open(out, 'a+') as fd:
            logger.info(
                "Writing {} documents' in {} into {}".format(
                    len(docs),
                    format,
                    out
                )
            )
            fd.write(outstring)
    else:
        print(outstring)



import re
import papis.downloaders.fallback


class Downloader(papis.downloaders.fallback.Downloader):

    def __init__(self, url):
        papis.downloaders.fallback.Downloader.__init__(self, url, name="aps")
        self.expected_document_extension = 'pdf'
        self.priority = 10

    @classmethod
    def match(cls, url):
        if re.match(r".*aps.org.*", url):
            return Downloader(url)
        else:
            return False

    def get_bibtex_url(self):
        url = self.uri
        burl = re.sub(r'/abstract', r'/export', url)\
            + "?type=bibtex&download=true"
        self.logger.debug("[bibtex url] = %s" % burl)
        return burl



import logging
import os
import papis.config
from papis.tui.app import Picker
from stevedore import extension
import papis.plugin

logger = logging.getLogger("pick")


def available_pickers():
    return pickers_mgr.entry_points_names()


def papis_pick(
        options, default_index=0,
        header_filter=lambda x: x, match_filter=lambda x: x
        ):
    if len(options) == 0:
        return ""
    if len(options) == 1:
        return options[0]

    picker = Picker(
        options,
        default_index,
        header_filter,
        match_filter
    )
    picker.run()
    return picker.options_list.get_selection()


pickers_mgr = extension.ExtensionManager(
    namespace='papis.picker',
    invoke_on_load=False,
    verify_requirements=True,
    propagate_map_exceptions=True,
    on_load_failure_callback=papis.plugin.stevedore_error_handler
)


def pick(
        options,
        default_index=0,
        header_filter=lambda x: x,
        match_filter=lambda x: x
        ):
    """Construct and start a :class:`Picker <Picker>`.
    """
    name = papis.config.get("picktool")
    try:
        picker = pickers_mgr[name].plugin
    except KeyError:
        logger.error("Invalid picker ({0})".format(name))
        logger.error(
            "Registered pickers are: {0}".format(available_pickers()))
    else:
        return picker(
            options,
            default_index=default_index,
            header_filter=header_filter,
            match_filter=match_filter
        )


def pick_doc(documents):
    """Pick a document from documents with the correct formatting

    :documents: List of documents
    :returns: Document

    """
    header_format_path = papis.config.get('header-format-file')
    if header_format_path is not None:
        with open(os.path.expanduser(header_format_path)) as fd:
            header_format = fd.read()
    else:
        header_format = papis.config.get("header-format")
    match_format = papis.config.get("match-format")
    pick_config = dict(
        header_filter=lambda x: papis.utils.format_doc(header_format, x),
        match_filter=lambda x: papis.utils.format_doc(match_format, x)
    )
    return pick(documents, **pick_config)



import click
import papis.document
import json
import logging


@click.command('json')
@click.pass_context
@click.argument('jsonfile', type=click.Path(exists=True))
@click.help_option('--help', '-h')
def explorer(ctx, jsonfile):
    """
    Import documents from a json file

    Examples of its usage are

    papis explore json lib.json pick

    """
    logger = logging.getLogger('explore:json')
    logger.info('Reading in json file {}'.format(jsonfile))
    docs = [
        papis.document.from_data(d) for d in json.load(open(jsonfile))
    ]
    ctx.obj['documents'] += docs
    logger.info('{} documents found'.format(len(docs)))



# -*- coding: utf-8 -*-
import subprocess
import multiprocessing
import time
import copy
from itertools import count, product
import os
import re
import papis.pick
import papis.config
import papis.commands
import papis.document
import papis.bibtex
import papis.exceptions
import logging
import papis.importer
import papis.downloaders
import colorama

logger = logging.getLogger("utils")
logger.debug("importing")


def general_open(fileName, key, default_opener=None, wait=True):
    """Wraper for openers


    >>> import tempfile; path = tempfile.mktemp()
    >>> general_open(path, 'nonexistent-key', wait=False).stdin == None
    True
    >>> general_open(path, 'nonexistent-key') > 0
    True
    >>> general_open([path], 'nonexistent-key', default_opener=lambda path: 42)
    42
    >>> general_open([path], 'nonexistent-key', default_opener=tempfile)
    Traceback (most recent call last):
    ...
    Warning: How should I use the opener ...?
    >>> papis.config.set('editor', 'echo')
    >>> general_open([path], 'editor', wait=False)
    <subprocess.Popen...>
    """
    try:
        opener = papis.config.get(key)
    except papis.exceptions.DefaultSettingValueMissing:
        if default_opener is None:
            default_opener = papis.config.get_default_opener()
        opener = default_opener
    if isinstance(fileName, list):
        fileName = papis.pick.pick(fileName)
    if isinstance(opener, str):
        import shlex
        cmd = shlex.split("{0} '{1}'".format(opener, fileName))
        logger.debug("cmd:  %s" % cmd)
        if wait:
            logger.debug("Waiting for process to finsih")
            return subprocess.call(cmd)
        else:
            logger.debug("Not waiting for process to finish")
            return subprocess.Popen(
                cmd, shell=False,
                stdin=None, stdout=None, stderr=None, close_fds=True
            )
    elif hasattr(opener, '__call__'):
        return opener(fileName)
    else:
        raise Warning("How should I use the opener %s?" % opener)


def open_file(file_path, wait=True):
    """Open file using the ``opentool`` key value as a program to
    handle file_path.

    :param file_path: File path to be handled.
    :type  file_path: str
    :param wait: Wait for the completion of the opener program to continue
    :type  wait: bool

    """
    general_open(fileName=file_path, key="opentool", wait=wait)


def format_doc(python_format, document, key=""):
    """Construct a string using a pythonic format string and a document.

    :param python_format: Python-like format string.
        (`see <
            https://docs.python.org/2/library/string.html#format-string-syntax
        >`_)
    :type  python_format: str
    :param document: Papis document
    :type  document: papis.document.Document
    :returns: Formated string
    :rtype: str
    """
    doc = key or papis.config.get("format-doc-name")
    fdoc = papis.document.Document()
    fdoc.update(document)
    try:
        return python_format.format(**{doc: fdoc})
    except Exception as e:
        return str(e)


def get_folders(folder):
    """This is the main indexing routine. It looks inside ``folder`` and crawls
    the whole directory structure in search for subfolders containing an info
    file.

    :param folder: Folder to look into.
    :type  folder: str
    :returns: List of folders containing an info file.
    :rtype: list
    """
    logger.debug("Indexing folders in '{0}'".format(folder))
    folders = list()
    for root, dirnames, filenames in os.walk(folder):
        if os.path.exists(os.path.join(root, papis.config.get('info-name'))):
            folders.append(root)
    logger.debug("{0} valid folders retrieved".format(len(folders)))
    return folders


def create_identifier(input_list):
    """This creates a generator object capable of iterating over lists to
    create combinations of that list that result in unique strings.
    Ideally for use in modifying an existing string to make it unique.

    Example:
    >>> import string
    >>> m = create_identifier(string.ascii_lowercase)
    >>> next(m)
    'a'

    (`see <
        https://stackoverflow.com/questions/14381940/
        >`_)

    :param input_list: list to iterate over
    :type  input_list: list

    """
    for n in count(1):
        for s in product(input_list, repeat=n):
            yield ''.join(s)


def confirm(prompt, yes=True, bottom_toolbar=None):
    """Confirm with user input

    :param prompt: Question or text that the user gets.
    :type  prompt: str
    :param yes: If yes should be the default.
    :type  yes: bool
    :returns: True if go ahead, False if stop
    :rtype:  bool

    """
    result = papis.utils.input(
        prompt,
        bottom_toolbar=bottom_toolbar,
        default='Y/n' if yes else 'y/N',
        validator_function=lambda x: x in 'YyNn',
        dirty_message='Please, write either "y" or "n" to confirm'
    )
    if yes:
        return result not in 'Nn'
    else:
        return result in 'Yy'


def text_area(title, text, lexer_name="", height=10, full_screen=False):
    """
    Small implementation of an editor/pager for small pieces of text.

    :param title: Title of the text_area
    :type  title: str
    :param text: Editable text
    :type  text: str
    :param lexer_name: If the editable text should be highlighted with
        some kind of grammar, examples are ``yaml``, ``python`` ...
    :type  lexer_name: str
    :param height: Max height of the text area
    :type  height: int
    :param full_screen: Wether or not the text area should be full screen.
    :type  full_screen: bool
    """
    from prompt_toolkit import Application
    from prompt_toolkit.enums import EditingMode
    from prompt_toolkit.buffer import Buffer
    from prompt_toolkit.layout.containers import HSplit, Window, WindowAlign
    from prompt_toolkit.layout.controls import (
        BufferControl, FormattedTextControl
    )
    from prompt_toolkit.layout.layout import Layout
    from prompt_toolkit.layout import Dimension
    from prompt_toolkit.key_binding import KeyBindings
    from prompt_toolkit.lexers import PygmentsLexer
    from pygments.lexers import find_lexer_class_by_name
    assert(type(title) == str)
    assert(type(text) == str)
    assert(type(lexer_name) == str)
    assert(type(height) == int)
    assert(type(full_screen) == bool)

    kb = KeyBindings()
    buffer1 = Buffer()
    buffer1.text = text

    @kb.add('c-q')
    def exit_(event):
        event.app.exit(0)

    @kb.add('c-s')
    def save_(event):
        event.app.return_text = buffer1.text

    class App(Application):
        return_text = None

    text_height = Dimension(min=0, max=height) if height is not None else None

    pygment_lexer = find_lexer_class_by_name(lexer_name)
    lexer = PygmentsLexer(pygment_lexer)
    text_window = Window(
        height=text_height,
        style='bg:black fg:ansiwhite',
        content=BufferControl(buffer=buffer1, lexer=lexer)
    )

    root_container = HSplit([
        Window(
            align=WindowAlign.LEFT,
            style='bg:ansiwhite',
            height=1,
            content=FormattedTextControl(
                text=[('fg:ansiblack bg:ansiwhite', title)]
            ),
            always_hide_cursor=True
        ),

        text_window,

        Window(
            height=1,
            width=None,
            align=WindowAlign.LEFT,
            style='bg:ansiwhite',
            content=FormattedTextControl(
                text=[(
                    'fg:ansiblack bg:ansiwhite',
                    "Quit [Ctrl-q]  Save [Ctrl-s]"
                )]
            )
        ),
    ])

    layout = Layout(root_container)

    layout.focus(text_window)

    app = App(
        editing_mode=(
            EditingMode.EMACS
            if papis.config.get('editmode', section='tui') == 'emacs'
            else EditingMode.VI
        ), layout=layout, key_bindings=kb, full_screen=full_screen
    )
    app.run()
    return app.return_text


def yes_no_dialog(title, text):
    from prompt_toolkit.shortcuts import yes_no_dialog
    from prompt_toolkit.styles import Style

    example_style = Style.from_dict({
        'dialog': 'bg:#88ff88',
        'dialog frame-label': 'bg:#ffffff #000000',
        'dialog.body': 'bg:#000000 #00ff00',
        'dialog shadow': 'bg:#00aa00',
    })

    return yes_no_dialog(
        title=title,
        text=text,
        style=example_style
    )


def input(
        prompt, default="", bottom_toolbar=None, multiline=False,
        validator_function=None, dirty_message=""):
    """Prompt user for input

    :param prompt: Question or text that the user gets.
    :type  prompt: str
    :param default: Default value to give if the user does not input anything
    :type  default: str
    :returns: User input or default
    :rtype:  bool

    """
    import prompt_toolkit
    import prompt_toolkit.validation
    if validator_function is not None:
        validator = prompt_toolkit.validation.Validator.from_callable(
            validator_function,
            error_message=dirty_message,
            move_cursor_to_end=True
        )
    else:
        validator = None
    if isinstance(prompt, str):
        fragments = [
            ('', prompt),
            ('fg:red', ' ({0})'.format(default)),
            ('', ': '),
        ]
    else:
        fragments = prompt

    result = prompt_toolkit.prompt(
        fragments,
        validator=validator,
        multiline=multiline,
        bottom_toolbar=bottom_toolbar,
        validate_while_typing=True
    )

    return result if result else default


def update_doc_from_data_interactively(document, data, data_name):
    import papis.tui.widgets.diff
    docdata = copy.copy(document)
    # do not compare some entries
    docdata.pop('files', None)
    docdata.pop('tags', None)
    document.update(
        papis.tui.widgets.diff.diffdict(
            docdata,
            data,
            namea=papis.document.describe(document), nameb=data_name))


def clean_document_name(doc_path):
    """Get a file path and return the basename of the path cleaned.

    It will also turn chinese, french, russian etc into ascii characters.

    :param doc_path: Path of a document.
    :type  doc_path: str
    :returns: Basename of the path cleaned
    :rtype:  str

    """
    import slugify
    regex_pattern = r'[^a-z0-9.]+'
    return slugify.slugify(
        os.path.basename(doc_path),
        word_boundary=True,
        regex_pattern=regex_pattern
    )


def locate_document_in_lib(document, library=None):
    """Try to figure out if a document is already in a library

    :param document: Document to be searched for
    :type  document: papis.document.Document
    :param library: Name of a valid papis library
    :type  library: str
    :returns: Document in library if found
    :rtype:  papis.document.Document
    :raises IndexError: Whenever document is not found in the library
    """
    db = papis.database.get(library=library)
    comparing_keys = eval(papis.config.get('unique-document-keys'))

    for k in comparing_keys:
        if not document.has(k):
            continue
        docs = db.query_dict({k: document[k]})
        if docs:
            return docs[0]

    raise IndexError("Document not found in library")


def locate_document(document, documents):
    """Try to figure out if a document is already within a list of documents.

    :param document: Document to be searched for
    :type  document: papis.document.Document
    :param documents: Documents to search in
    :type  documents: list
    :returns: papis document if it is found

    """
    # if these keys exist in the documents, then check those first
    # TODO: find a way to really match well titles and author
    comparing_keys = eval(papis.config.get('unique-document-keys'))
    for d in documents:
        for key in comparing_keys:
            if key in document.keys() and key in d.keys():
                if re.match(document[key], d[key], re.I):
                    return d


def get_document_extension(document_path):
    """Get document extension

    :document_path: Path of the document
    :returns: Extension (string)

    """
    import filetype
    filetype.guess(document_path)
    kind = filetype.guess(document_path)
    if kind is None:
        m = re.match(r"^.*\.([^.]+)$", os.path.basename(document_path))
        return m.group(1) if m else 'data'
    else:
        return kind.extension


def folders_to_documents(folders):
    """Turn folders into documents, this is done in a multiprocessing way, this
    step is quite critical for performance.

    :param folders: List of folder paths.
    :type  folders: list
    :returns: List of document objects.
    :rtype:  list
    """
    logger = logging.getLogger("utils:dir2doc")
    np = multiprocessing.cpu_count()
    logger.debug("converting folder into documents on {0} cores".format(np))
    pool = multiprocessing.Pool(np)
    begin_t = time.time()
    result = pool.map(papis.document.from_folder, folders)
    pool.close()
    pool.join()
    logger.debug("done in %.1f ms" % (1000*time.time()-1000*begin_t))
    return result


def get_cache_home():
    """Get folder where the cache files are stored, it retrieves the
    ``cache-dir`` configuration setting. It is ``XDG`` standard compatible.

    :returns: Full path for cache main folder
    :rtype:  str

    """
    user_defined = papis.config.get('cache-dir')
    if user_defined is not None:
        path = os.path.expanduser(user_defined)
    else:
        path = os.path.expanduser(
            os.path.join(os.environ.get('XDG_CACHE_HOME'), 'papis')
        ) if os.environ.get(
            'XDG_CACHE_HOME'
        ) else os.path.expanduser(
            os.path.join('~', '.cache', 'papis')
        )
    if not os.path.exists(path):
        os.makedirs(path)
    return path


def geturl(url):
    """Quick and dirty file get request utility.
    """
    assert(isinstance(url, str))
    import requests
    session = requests.Session()
    session.headers = {'User-Agent': papis.config.get('user-agent')}
    return session.get(url).content


def get_matching_importer_or_downloader(matching_string):
    importers = []
    logger = logging.getLogger("utils:matcher")
    for importer_cls in (papis.importer.get_importers() +
                         papis.downloaders.get_downloaders()):
        importer = importer_cls.match(matching_string)
        logger.debug(
            "trying with importer {c.Back.BLACK}{c.Fore.YELLOW}{name}"
            "{c.Style.RESET_ALL}".format(
                c=colorama, name=importer_cls))
        if importer:
            logger.info(
                "{f} {c.Back.BLACK}{c.Fore.GREEN}matches {name}"
                "{c.Style.RESET_ALL}".format(
                    f=matching_string, c=colorama, name=importer.name))
            try:
                importer.fetch()
            except Exception as e:
                logger.error(e)
            else:
                importers.append(importer)
    return importers



import logging
import papis.config
import papis.utils
import doi
import habanero
import re
import os
import click
import papis.document
import papis.importer
import papis.downloaders.base
import tempfile
import collections

logger = logging.getLogger("crossref")
logger.debug("importing")


_filter_names = set([
    "has_funder", "funder", "location", "prefix", "member", "from_index_date",
    "until_index_date", "from_deposit_date", "until_deposit_date",
    "from_update_date", "until_update_date", "from_created_date",
    "until_created_date", "from_pub_date", "until_pub_date",
    "from_online_pub_date", "until_online_pub_date", "from_print_pub_date",
    "until_print_pub_date", "from_posted_date", "until_posted_date",
    "from_accepted_date", "until_accepted_date", "has_license",
    "license_url", "license_version", "license_delay",
    "has_full_text", "full_text_version", "full_text_type",
    "full_text_application", "has_references", "has_archive",
    "archive", "has_orcid", "has_authenticated_orcid",
    "orcid", "issn", "type", "directory", "doi", "updates", "is_update",
    "has_update_policy", "container_title", "category_name", "type",
    "type_name", "award_number", "award_funder", "has_assertion",
    "assertion_group", "assertion", "has_affiliation", "alternative_id",
    "article_number", "has_abstract", "has_clinical_trial_number",
    "content_domain", "has_content_domain", "has_crossmark_restriction",
    "has_relation", "relation_type", "relation_object", "relation_object_type",
    "public_references", "publisher_name", "affiliation",
])

_types_values = [
    "book-section", "monograph", "report", "peer-review", "book-track",
    "journal-article", "book-part", "other", "book", "journal-volume",
    "book-set", "reference-entry", "proceedings-article", "journal",
    "component", "book-chapter", "proceedings-series", "report-series",
    "proceedings", "standard", "reference-book", "posted-content",
    "journal-issue", "dissertation", "dataset", "book-series", "edited-book",
    "standard-series",
]

_sort_values = [
"relevance", "score", "updated", "deposited", "indexed", "published",
"published-print", "published-online", "issued", "is-referenced-by-count",
"references-count",
]


_order_values = ['asc', 'desc']


type_converter = {
    "book": "book",
    "book-chapter": "inbook",
    "book-part": "inbook",
    "book-section": "inbook",
    "book-series": "incollection",
    "book-set": "incollection",
    "book-track": "inbook",
    "dataset": "misc",
    "dissertation": "phdthesis",
    "edited-book": "book",
    "journal-article": "article",
    "journal-issue": "misc",
    "journal-volume": "article",
    "monograph": "monograph",
    "other": "misc",
    "peer-review": "article",
    "posted-content": "misc",
    "proceedings-article": "inproceedings",
    "proceedings": "inproceedings",
    "proceedings-series": "inproceedings",
    "reference-book": "book",
    "report": "report",
    "report-series": "inproceedings",
    "standard-series": "incollection",
    "standard": "techreport",
}


key_conversion = collections.OrderedDict({
    "DOI": {"key": "doi"},
    "URL": {"key": "url"},
    "author": {
        "key": "author_list",
        "action": lambda authors: [
            {k: a.get(k) for k in ['given', 'family', 'affiliation']}
            for a in authors
        ],
    },
    "container-title": {"key": "journal", "action": lambda x: x[0]},
    "issue": {},
    # "issued": {"key": "",},
    "language": {},
    "abstract": {},
    "ISBN": {"key": "isbn"},
    "page": {
        "key": "pages",
        "action": lambda p: re.sub(r"(-[^-])", r"-\1", p),
    },
    "link": {
        "key": papis.config.get('doc-url-key-name'),
        "action": lambda x: x[1]["URL"]
    },
    "issued": [
        {"key": "year", "action": lambda x: x.get("date-parts")[0][0]},
        {"key": "month", "action": lambda x: x.get("date-parts")[0][1]}
    ],
    "published-online": [
        {"key": "year", "action": lambda x: x.get("date-parts")[0][0]},
        {"key": "month", "action": lambda x: x.get("date-parts")[0][1]}
    ],
    "published-print": [
        {"key": "year", "action": lambda x: x.get("date-parts")[0][0]},
        {"key": "month", "action": lambda x: x.get("date-parts")[0][1]}
    ],
    "publisher": {},
    "reference": {
        "key": "citations",
        "action": lambda cs: [
            {key.lower(): c[key]
                for key in set(c.keys()) - set(("key", "doi-asserted-by"))}
            for c in cs
        ]
    },
    # "short-title": { "key": "", },
    # "subtitle": { "key": "", },
    "title": {"action": lambda t: " ".join(t)},
    "type": {"action": lambda t: type_converter[t]},
    "volume": {},
    "event": [  # Conferences
        {"key": "venue", "action": lambda x: x["location"]},
        {"key": "booktitle", "action": lambda x: x["name"]},
        {"key": "year", "action": lambda x: x['start']["date-parts"][0][0]},
        {"key": "month", "action": lambda x: x['start']["date-parts"][0][1]},
    ],
})


def crossref_data_to_papis_data(data):
    global key_conversion
    new_data = papis.document.keyconversion_to_data(key_conversion, data)
    new_data['author'] = papis.document.author_list_to_author(new_data)
    return new_data


def _get_crossref_works(**kwargs):
    cr = habanero.Crossref()
    return cr.works(**kwargs)


def get_data(query="", author="", title="", dois=[], max_results=0,
        filters={}, sort="score", order='desc'):
    global _filter_names
    global _sort_values
    assert(isinstance(dois, list)), 'dois must be a list'
    assert(sort in _sort_values), 'Sort value not valid'
    assert(order in _order_values), 'Sort value not valid'
    assert(isinstance(filters, dict)), 'filters must be a dictionary'
    if filters:
        if not set(filters.keys()) & _filter_names == set(filters.keys()):
            raise Exception(
                'Filter keys must be one of {0}'
                .format(','.join(_filter_names))
            )
    data = dict(
        query=query, query_author=author,
        ids=dois,
        query_title=title, limit=max_results
    )
    kwargs = {key: data[key] for key in data.keys() if data[key]}
    if not dois:
        kwargs.update(dict(sort=sort, order=order))
    try:
        results = _get_crossref_works(filter=filters, **kwargs)
    except Exception as e:
        logger.error(e)
        return []

    if isinstance(results, list):
        docs = [d["message"] for d in results]
    elif isinstance(results, dict):
        if 'message' not in results.keys():
            logger.error(
                "Error retrieving from xref, I got an incorrect message")
            return []
        message = results['message']
        if "items" in message.keys():
            docs = message['items']
        else:
            docs = [message]
    else:
        logger.error("Error retrieving from xref, I got an incorrect message")
        return []
    logger.debug("Retrieved {} documents".format(len(docs)))
    return [
        crossref_data_to_papis_data(d)
        for d in docs
    ]


def doi_to_data(doi_string):
    """Search through crossref and get a dictionary containing the data

    :param doi_string: Doi identificator or an url with some doi
    :type  doi_string: str
    :returns: Dictionary containing the data
    :raises ValueError: If no data could be retrieved for the doi

    """
    global logger
    assert(isinstance(doi_string, str))
    doi_string = doi.get_clean_doi(doi_string)
    results = get_data(dois=[doi_string])
    if results:
        return results[0]
    else:
        raise ValueError(
            "Couldn't get data for doi ({doi})".format(doi=doi_string)
        )


@click.command('crossref')
@click.pass_context
@click.help_option('--help', '-h')
@click.option('--query', '-q', help='General query', default=None)
@click.option('--author', '-a', help='Author of the query', default=None)
@click.option('--title', '-t', help='Title of the query', default=None)
@click.option('--max', '-m', help='Maximum number of results', default=20)
@click.option(
    '--filter', '-f', help='Filters to apply', default=(),
    type=(click.Choice(_filter_names), str),
    multiple=True
)
@click.option(
    '--order', '-o', help='Order of appearance according to sorting',
    default='desc', type=click.Choice(_order_values), show_default=True
)
@click.option(
    '--sort', '-s', help='Sorting parameter', default='score',
    type=click.Choice(_sort_values), show_default=True
)
def explorer(ctx, query, author, title, max, filter, sort, order):
    """
    Look for documents on crossref.org.

    Examples of its usage are

    papis explore crossref -a 'Albert einstein' pick export --bibtex lib.bib

    """
    logger = logging.getLogger('explore:crossref')
    logger.info('Looking up...')
    data = get_data(
        query=query,
        author=author,
        title=title,
        max_results=max,
        filters=dict(filter),
        sort=sort,
        order=order
    )
    docs = [papis.document.from_data(data=d) for d in data]
    ctx.obj['documents'] += docs
    logger.info('{} documents found'.format(len(docs)))


class DoiFromPdfImporter(papis.importer.Importer):

    """Importer parsing a doi from a pdf file"""

    def __init__(self, **kwargs):
        papis.importer.Importer.__init__(self, name='pdf2doi', **kwargs)
        self.doi = None

    @classmethod
    def match(cls, uri):
        if (os.path.isdir(uri) or not os.path.exists(uri) or
                not papis.utils.get_document_extension(uri) == 'pdf'):
            return None
        importer = DoiFromPdfImporter(uri=uri)
        importer.doi = doi.pdf_to_doi(uri, maxlines=2000)
        return importer if importer.doi else None

    def fetch(self):
        self.logger.info("Trying to parse doi from file {0}".format(self.uri))
        if not self.doi:
            self.doi = doi.pdf_to_doi(self.uri, maxlines=2000)
        if self.doi:
            self.logger.info("Parsed doi {0}".format(self.doi))
            self.logger.warning(
                "There is no guarantee that this doi is the one")
            importer = Importer(uri=self.doi)
            importer.fetch()
            self.ctx = importer.ctx


class Importer(papis.importer.Importer):

    """Importer getting files and data form a doi through crossref.org"""

    def __init__(self, **kwargs):
        papis.importer.Importer.__init__(self, name='doi', **kwargs)

    @classmethod
    def match(cls, uri):
        try:
            doi.validate_doi(uri)
        except ValueError:
            return None
        else:
            return Importer(uri=uri)

    @classmethod
    def match_data(cls, data):
        if 'doi' in data:
            return Importer(uri=data['doi'])

    def fetch(self):
        self.logger.info("using doi {0}".format(self.uri))
        doidata = papis.crossref.get_data(dois=[self.uri])
        if doidata:
            self.ctx.data = doidata[0]
            if papis.config.get('doc-url-key-name') in self.ctx.data.keys():
                doc_url = self.ctx.data[papis.config.get('doc-url-key-name')]
                self.logger.info(
                    'trying to download document from {0}..'
                    .format(doc_url))
                document_data = papis.utils.geturl(doc_url)
                tmp_filepath = tempfile.mktemp()
                self.logger.debug("Saving in %s" % tmp_filepath)
                with open(tmp_filepath, 'wb+') as fd:
                    fd.write(document_data)
                self.ctx.files.append(tmp_filepath)


class FromCrossrefImporter(papis.importer.Importer):

    """Importer that gets data from querying to crossref"""

    def __init__(self, **kwargs):
        papis.importer.Importer.__init__(self, name='crossref', **kwargs)

    @classmethod
    def match(cls, uri):
        # There is no way to check if it matches
        return None

    @classmethod
    def match_data(cls, data):
        if 'title' in data:
            return FromCrossrefImporter(uri=data['title'])

    def fetch(self):
        self.logger.info("querying '{0}' to crossref.org".format(self.uri))
        docs = [
            papis.document.from_data(d)
            for d in get_data(query=self.uri)
        ]
        if docs:
            self.logger.info("got {0} matches, picking...".format(len(docs)))
            doc = papis.pick.pick_doc(docs)
            if doc:
                importer = Importer(uri=doc['doi'])
                importer.fetch()
                self.ctx = importer.ctx


class Downloader(papis.downloaders.base.Downloader):

    def __init__(self, uri):
        papis.downloaders.base.Downloader.__init__(self, uri=uri, name="doi")

    @classmethod
    def match(cls, uri):
        if doi.find_doi_in_text(uri):
            return Downloader(uri)
        else:
            return None

    def fetch(self):
        importer = Importer(uri=doi.find_doi_in_text(self.uri))
        importer.fetch()
        self.ctx = importer.ctx
